#!/usr/bin/env python

import sys
import pprint

import radical.pilot as rp
import radical.analytics as ra

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'

"""
This example illustrates the use of the method ra.Session.duration()
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # and here we go. Once we filter our session object so to keep only the
    # relevent entities (as seen in example 03), we are ready to perform our
    # analyses :) Currently, RADICAL-Analytics supports two types of analysis:
    # duration and concurrency. This examples shows how to use the RA API to
    # performan duration analysis, using both states and events.
    #
    # First we look at the state model of our session. We saw how to print it
    # out in example 00:
    ppheader("state models")
    pprint.pprint(session.describe('state_model'))

    # Let's say that we want to see for how long all the pilot(s) we use have
    # been active. Looking at the state model of the entity of type 'pilot' and
    # to the documentation of RADICAL-Pilot, we know that a pilot is active
    # between the state 'ACTIVE' and one of the three final states 'DONE',
    # 'CANCELED', 'FAILED' of all the entities of RP.
    ppheader("Time spent by the pilots being active")
    pilots = session.filter(etype='pilot', inplace=False)
    durations = pilots.duration([rp.PMGR_ACTIVE, rp.FINAL])
    pprint.pprint(durations)

    # Now, we want to do the same for the all the entities of type 'unit':
    ppheader("Time spent by the units being active")
    units = session.filter(etype='unit', inplace=False)
    duration_active = units.duration([rp.AGENT_EXECUTING, rp.FINAL])
    pprint.pprint(duration_active)

    conc = units.concurrency(state=[rp.AGENT_EXECUTING,
                                    rp.AGENT_STAGING_OUTPUT_PENDING],
                                    sampling=2.0)
    pprint.pprint(conc)

    # The careful reader will have noticed that the previous duration includes
    # the time spent by the units to execute and to stage data out. We can
    # separate the two by using the state 'AGENT_STAGING_OUTPUT_PENDING',
    # instead of the final states:
    ppheader("Time spent by the units executing their kernel")
    units = session.filter(etype='unit', inplace=False)
    duration_exec = units.duration([rp.AGENT_EXECUTING, rp.AGENT_STAGING_OUTPUT_PENDING])
    pprint.pprint(duration_exec)

    # We calculate the time spent doing staging out by all entities of type
    # 'unit' with 'AGENT_STAGING_OUTPUT_PENDING' and the final states:
    ppheader("Time spent by the units performing staging out")
    units = session.filter(etype='unit', state=rp.DONE, inplace=False)
    duration_sout = units.duration([rp.AGENT_STAGING_OUTPUT_PENDING, rp.FINAL])
    pprint.pprint(duration_sout)

    # we print the timestamps for the units for when they entered certain states
    ppheader("Timestamps for state transitions")
    print('[rp.AGENT_STAGING_OUTPUT_PENDING]:')
    pprint.pprint(units.timestamps(state=[rp.AGENT_STAGING_OUTPUT_PENDING]))
    print('rp.AGENT_EXECUTING')
    pprint.pprint(units.timestamps(state=rp.AGENT_EXECUTING))
    print('[rp.AGENT_EXECUTING, rp.AGENT_STAGING_OUTPUT_PENDING]')
    pprint.pprint(units.timestamps(state=[rp.AGENT_EXECUTING, rp.AGENT_STAGING_OUTPUT_PENDING]))

    print(""" The very careful reader may have noticed that the sum of the time
    spent by the units to execute their kernel and performing staging out may be
    greater than the time spent by the units being active. This is explained by
    the potential overlapping between the time spent executing the kernell and
    that spent staging out. This overlapping is accounted for when calculating
    the time spent by the units being active.""")

    ppheader("Ovelapping between executing and staging")
    overlap = (duration_exec + duration_sout) - duration_active
    print(overlap)

    sys.exit(0)



# pylint: disable=W0102,W0212

import re
import os
import sys
import copy
import tarfile

import pickle as pickle

import more_itertools as mit
import radical.utils  as ru

from .entity import Entity


# ------------------------------------------------------------------------------
#
class Session(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, src, stype, sid=None, _entities=None, _init=True):
        '''
        Create a radical.analytics session for analysis.

        The session is created from a set of traces, which usually have been
        produced by a Session object in the RCT stack, such as radical.pilot or
        radical.entk. Profiles are accepted in two forms: in a directory, or in
        a tarball (of such a directory).  In the latter case, the tarball are
        extracted into `$TMP`, and then handled just as the directory case.

        If no `sid` (session ID) is specified, that ID is derived from the
        directory name.

        '''

        if _init:
            # if no sid is given, derive it from the src path
            sid, src, tgt, ext = self._get_sid(sid, src)

        else:
            assert sid
            assert src
            tgt = None

        self._sid   = sid
        self._src   = src
        self._stype = stype

        if tgt and not os.path.exists(tgt):

            # need to extract
            print(('extract tarball to %s' % tgt))
            try:
                if ext in ['tbz', 'tar.bz', 'tbz2', 'tar.bz2']:
                    tf = tarfile.open(name=src, mode='r:bz2')
                    tf.extractall(path=os.path.dirname(tgt))
                elif ext in ['tgz, tar.gz']:
                    tf = tarfile.open(name=src, mode='r:gz')
                    tf.extractall(path=os.path.dirname(tgt))
                else:
                    raise ValueError('cannot handle extension %s' % ext)

            except Exception as e:
                raise RuntimeError(
                    'Cannot extract tarball: %s' % repr(e)) from e

        if tgt:
            self._src = tgt

      # print 'sid: %s [%s]' % (sid, stype)
      # print 'src: %s'      % src

        if stype == 'radical':

            # src is expected to point either to a single profile, or to
            # a directory tree containing profiles
            if not src:
                raise ValueError('RA session types need `src` specified')

            profiles = list()
            if os.path.isfile(src):
                profiles.append(src)
            else:
                def _walkdir(path, profiles=[]):
                    for root, dirs, files in os.walk(path):
                        for f in files:
                            if f.endswith('.prof'):
                                profiles.append('%s/%s' % (root, f))
                        for d in dirs:
                            _walkdir('%s/%s' % (path, d), profiles)
                        return profiles
                profiles = _walkdir(src)

            profiles                = ru.read_profiles(profiles, sid=sid)
            self._profile, accuracy = ru.combine_profiles(profiles)
            self._description       = {'tree'     : dict(),
                                       'entities' : list(),
                                       'hostmap'  : dict(),
                                       'accuracy' : 0.0}

        elif stype == 'radical.pilot':

            try:
                import radical.pilot.utils as rpu
            except Exception as e:
                raise RuntimeError('radical.analytics requires the '
                                   'radical.pilot module to analyze this '
                                   'session - please install it.') from e

            self._profile, accuracy, hostmap = \
                    rpu.get_session_profile(sid=sid, src=self._src)
            self._description = \
                    rpu.get_session_description(sid=sid, src=self._src)

            self._description['accuracy'] = accuracy
            self._description['hostmap']  = hostmap


        elif stype == 'radical.entk':

            try:
                import radical.entk.utils as reu
            except Exception as e:
                raise RuntimeError('radical.analytics requires the '
                                   'radical.entk module to analyze this '
                                   'session - please install it.') from e

            self._profile, accuracy, hostmap \
                              = reu.get_session_profile    (sid=sid, src=self._src)
            self._description = reu.get_session_description(sid=sid, src=self._src)

            self._description['accuracy'] = accuracy
            self._description['hostmap']  = hostmap


        else:
            raise ValueError('unsupported session type [%s]' % stype)


        self._t_start = None
        self._t_stop  = None
        self._ttc     = None
        self._log     = ru.Logger('radical.analytics')
        self._rep     = ru.Reporter('radical.analytics')


        # user defined time offset
        self._tzero   = 0.0

        # internal state is represented by a dict of entities:
        # dict keys are entity uids (which are assumed to be unique per
        # session), dict values are ra.Entity instances.
        self._entities = dict()
        if _init:
            self._initialize_entities(self._profile)

        # we do some bookkeeping in self._properties where we keep a list of
        # property values around which we encountered in self._entities.
        self._properties = dict()
        if _init:
            self._initialize_properties()

      # print('session loaded')

        # FIXME: we should do a sanity check that all encountered states and
        #        events are part of the respective state and event models
      # self.consistency()


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _get_sid(sid, src):

        tgt = None
        ext = None

        if not os.path.exists(src):
            raise ValueError('src [%s] does not exist' % src)

        if os.path.isdir(src):
            pass

        elif os.path.isfile(src):

            # src is a file - we assume its a tarball and extract it
            if  src.endswith('.prof'):
                # use as is
                tgt = src

            elif src.endswith('.tgz') or \
                 src.endswith('.tbz')    :
                tgt = src[:-4]
                ext = src[-3:]

            elif src.endswith('.tbz2'):
                tgt = src[:-5]
                ext = src[-4:]

            elif src.endswith('.tar.gz') or \
                 src.endswith('.tar.bz')    :
                tgt = src[:-7]
                ext = src[-6:]

            elif src.endswith('.tar.bz2'):
                tgt = src[:-8]
                ext = src[-7:]

            elif src.endswith('.prof'):
                tgt = None
                ext = None

            else:
                raise ValueError('src does not look like a tarball or profile')

        if not sid:

            if tgt: to_check = tgt
            else  : to_check = src

            if to_check.endswith('/'):
                to_check = src[:-1]
            sid = os.path.basename(to_check)

        return sid, src, tgt, ext


    # --------------------------------------------------------------------------
    #
    def __getstate__(self):

        state = {
                 'sid'         : self._sid,
                 'src'         : self._src,
                 'stype'       : self._stype,
                 'profile'     : self._profile,
                 'description' : self._description,

                 't_start'     : self._t_start,
                 't_stop'      : self._t_stop,
                 'ttc'         : self._ttc,

                 'entities'    : self._entities,
                 'properties'  : self._properties,
                }

        return state


    # --------------------------------------------------------------------------
    #
    def __setstate__(self, state):

        self._sid         = state['sid']
        self._src         = state['src']
        self._stype       = state['stype']
        self._profile     = state['profile']
        self._description = state['description']

        self._t_start     = state['t_start']
        self._t_stop      = state['t_stop']
        self._ttc         = state['ttc']

        self._entities    = state['entities']
        self._properties  = state['properties']

        self._log         = ru.Logger('radical.analytics')
        self._rep         = ru.Reporter('radical.analytics')


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def create(src, stype, sid=None, _entities=None, _init=True, cache=True):

        sid, src, _, _ = Session._get_sid(sid, src)
        base  = ru.get_radical_base('radical.analytics.cache')
        cache = '%s/%s.pickle' % (base, sid)

        if _entities or not cache:
            # no caching
            session = Session(src, stype, sid, _entities, _init)

        try:
            with open(cache, 'rb') as fin:
                data = fin.read()
                session = pickle.loads(data)
                print(('using cache for %s' % sid))
         #      import pprint
         #      j = ru.read_json("%s/%s.json" % (src, sid))
         #      rd = j['pilot'][0]['resource_details']
         #      session.get(etype='pilot')[0].cfg['resource_details'] = rd
         #      pprint.pprint(session.get(etype='pilot')[0].cfg)
         #  with open(cache, 'wb') as fout:
         #    # session = Session(src, stype, sid, _entities, _init)
         #      fout.write(pickle.dumps(session, protocol=pickle.HIGHEST_PROTOCOL))

        except Exception:
            print(('no cache for %s' % sid))
            with open(cache, 'wb') as fout:
                session = Session(src, stype, sid, _entities, _init)
                fout.write(pickle.dumps(session, protocol=pickle.HIGHEST_PROTOCOL))

        return session


    # --------------------------------------------------------------------------
    #
    def __deepcopy___(self, memo):

        cls = self.__class__
        ret = cls(sid=self._sid, src=None, stype=self._stype, _init=False)

        memo[id(self)] = ret

        for k, v in list(self.__dict__.items()):
            setattr(ret, k, copy.deepcopy(v, memo))

        return ret


    # --------------------------------------------------------------------------
    #
    def _reinit(self, entities):
        '''
        After creating a session clone, we have identical sets of descriptions,
        profiles, and entities.  However, if we apply a filter during the clone
        creation, we end up with a deep copy which should have a **different**
        set of entities.  This method applies that new entity set to such a
        cloned session.
        '''

        self._entities = entities

        # FIXME: we may want to filter the session description etc. wrt. to the
        #        entity types remaining after a filter.


    # --------------------------------------------------------------------------
    #
    @property
    def t_start(self):
        return self._t_start

    @property
    def t_stop(self):
        return self._t_stop

    @property
    def ttc(self):
        return self._ttc

    @property
    def t_range(self):
        return [self._t_start, self._t_stop]

    @property
    def uid(self):
        return self._sid

    @property
    def stype(self):
        return self._stype


    # --------------------------------------------------------------------------
    #
    def _initialize_entities(self, profile):
        '''
        Populates self._entities from profile and self._description.

        NOTE: We derive entity types via some heuristics for now: we assume the
        first part of any dot-separated uid to signify an entity type.
        '''

        # create entities from the profile events:
        entity_events = dict()

        for event in profile:

            if event[ru.TIME] < -1:  # allow fow 1sec rounding error
                raise ValueError('invalid time stamp: %s' % event)

            uid = event[ru.UID]
            if uid not in entity_events:
                entity_events[uid] = list()
            entity_events[uid].append(event)

        # for all uids found,  create and store an entity.  We look up the
        # entity type in one of the events (and assume it is consistent over
        # all events for that uid)
        for uid,events in list(entity_events.items()):
            details = self._description.get('tree', dict()).get(uid, dict())

            # hostid should be handled on RP level
            hostid  = self._description.get('hostmap', dict()).get(uid)
            if hostid:
                details['hostid'] = hostid

            self._entities[uid] = Entity(_uid=uid,
                                         _profile=events,
                                         _details=details)


    # --------------------------------------------------------------------------
    #
    def _initialize_properties(self):
        '''
        populate `self._properties` from `self._entities`.  `self._properties`∫
        has the following format::

            {
              'state' : {
                'NEW'      : 10,
                'RUNNING'  :  8,
                'DONE      :  7,
                'FAILED'   :  1,
                'CANCELED' :  2
              }
            }

        We count how often any property value appears in the current set of
        entities.

        RA knows exactly 4 properties:
          - uid   (entity identifiers)
          - etype (type of entities)
          - event (names of events)
          - state (state identifiers)
        '''

        # FIXME: initializing properties can be expensive, and we might not
        #        always need them anyway.  So we can lazily defer this
        #        initialization stop until the first query which requires them.

        # we do *not* look at profile and descriptions anymore, those are only
        # evaluated once on construction, in `_initialize_entities()`.  Now we
        # don't parse all that stuff again, but only re-initialize after
        # in-place filtering etc.
        self._properties = {'uid'   : dict(),
                            'etype' : dict(),
                            'event' : dict(),
                            'state' : dict()}

        if self._entities:
            self._t_start = sys.float_info.max
            self._t_stop  = sys.float_info.min

        for euid,e in list(self._entities.items()):

            self._t_start = min(self._t_start, e.t_start)
            self._t_stop  = max(self._t_stop,  e.t_stop )

            if euid in self._properties['uid']:
                raise RuntimeError('duplicated uid %s' % euid)
            self._properties['uid'][euid] = 1

            if e.etype not in self._properties['etype']:
                self._properties['etype'][e.etype] = 0
            self._properties['etype'][e.etype] += 1

            for state in e.states:
                if state not in self._properties['state']:
                    self._properties['state'][state] = 0
                self._properties['state'][state] += 1

            for event in e.events:
                name = event[ru.EVENT]
                if name not in self._properties['event']:
                    self._properties['event'][name] = 0
                self._properties['event'][name] += 1

        if self._entities:
            self._ttc = self._t_stop - self._t_start


    # --------------------------------------------------------------------------
    #
    def _apply_filter(self, etype=None, uid=None, name=None,
                            state=None, event=None, time=None):

        # iterate through all self._entities and collect UIDs of all entities
        # which match the given set of filters (after removing all events which
        # are not in the given time ranges)
        etype = ru.as_list(etype)
        uids  = ru.as_list(uid)
        names = ru.as_list(name)
        state = ru.as_list(state)
        event = ru.as_list(event)
        time  = ru.as_list(time )

        ret = list()
        for eid, entity in list(self._entities.items()):

            if etype and entity.etype not in etype: continue

            if uids:
                try:
                    re_pattern = re.Pattern
                except AttributeError:
                    re_pattern = None
                    self._log.warn('re.Pattern is not supported within this '
                                   'python version')

                keep = False
                for uid in uids:
                    if re_pattern and isinstance(uid, re_pattern):
                        # uid is actually a regex we use for matching
                        if uid.match(entity.uid):
                            keep = True
                            break
                    else:
                        # uid is a specific string to look out for
                        if entity.uid == uid:
                            keep = True
                            break

                if not keep:
                    continue

            if names:
                try:
                    re_pattern = re.Pattern
                except AttributeError:
                    re_pattern = None
                    self._log.warn('re.Pattern is not supported within this '
                                   'python version')

                keep = False
                for name in names:
                    if re_pattern and isinstance(name, re_pattern):
                        # name is actually a regex we use for matching
                        if name.match(entity.name):
                            keep = True
                            break
                    else:
                        # name is a specific string to look out for
                        if entity.name == name:
                            keep = True
                            break

                if not keep:
                    continue

            if state:
                match = False
                for s,stuple in list(entity.states.items()):
                    if time and not ru.in_range(stuple[ru.TIME], time):
                        continue
                    if s in state:
                        match = True
                        break
                if not match:
                    continue

            if event:
                match = False
                for etuple in entity.events:
                    if time and not ru.in_range(etuple[ru.TIME], time):
                        continue
                    if etuple[ru.EVENT] in event:
                        match = True
                        break
                if not match:
                    continue

            # all existing filters have been passed - this is a match!
            ret.append(eid)

        return ret


    # --------------------------------------------------------------------------
    #
    def _dump(self):

        for uid,entity in list(self._entities.items()):
            print(('\n\n === %s' % uid))
            entity.dump()
            for event in entity.events:
                print(('  = %s' % event))
                for e in entity.events[event]:
                    print(('    %s' % e))


    # --------------------------------------------------------------------------
    #
    def list(self, pname=None):

        if not pname:
            # return the name of all known properties
            return list(self._properties.keys())

        if isinstance(pname, list):
            return_list = True
            pnames = pname
        else:
            return_list = False
            pnames = [pname]

        ret = list()
        for _pname in pnames:
            if _pname not in self._properties:
                raise KeyError('no such property known (%s) / %s'
                        % (_pname, list(self._properties.keys())))
            ret.append(list(self._properties[_pname].keys()))

        if return_list: return ret
        else          : return ret[0]


    # --------------------------------------------------------------------------
    #
    def get(self, etype=None, uid=None, name=None,
                  state=None, event=None, time=None):

        uids = self._apply_filter(etype=etype, uid=uid, name=name,
                                  state=state, event=event, time=time)
        return [self._entities[_uid] for _uid in uids]


    # --------------------------------------------------------------------------
    #
    def filter(self, etype=None, uid=None, name=None,
                     state=None, event=None, time=None, inplace=True):

        uids = self._apply_filter(etype=etype, uid=uid, name=name,
                                  state=state, event=event, time=time)

        if inplace:
            # filter our own entity list, and refresh the entity based on
            # the new list
            if uids != list(self._entities.keys()):
                self._entities = {uid:self._entities[uid] for uid in uids}
                self._initialize_properties()
            return self

        else:
            # create a new session with the resulting entity list
            ret = Session(sid=self._sid, stype=self._stype, src=self._src,
                          _init=False)
            ret._reinit(entities={uid:self._entities[uid] for uid in uids})
            ret._initialize_properties()
            return ret


    # --------------------------------------------------------------------------
    #
    def describe(self, mode=None, etype=None):

        if mode not in [None, 'state_model', 'state_values',
                              'event_model', 'relations',
                              'statistics']:
            raise ValueError('describe parameter "mode" invalid')

        if not etype and not mode:
            # no entity filter applied: return the full description
            return self._description

        if mode == 'statistics':
            return self._properties

        if not etype:
            etype = self.list('etype')

        if not isinstance(etype,list):
            etype = [etype]

        ret = dict()
        for et in etype:

            state_model  = None
            state_values = None
            event_model  = None

            if et in self._description.get('entities', dict()):
                state_model  = self._description['entities'][et]['state_model']
                state_values = self._description['entities'][et]['state_values']
                event_model  = self._description['entities'][et]['event_model']

            if not state_model  : state_model  = dict()
            if not state_values : state_values = dict()
            if not event_model  : event_model  = dict()

            if not mode:
                ret[et] = {'state_model'  : state_model,
                           'state_values' : state_values,
                           'event_model'  : event_model}

            elif mode == 'state_model':
                ret[et] = {'state_model'  : state_model}

            elif mode == 'state_values':
                ret[et] = {'state_values' : state_values}

            elif mode == 'event_model':
                ret[et] = {'event_model'  : event_model}

        if not mode or mode == 'relations':
            if len(etype) != 2:
                raise ValueError('relations expect an etype *tuple*')

            # we interpret the query as follows: for the two given etypes, walk
            # through the relationship tree and for all entities of etype[0]
            # return a list of all child entities of etype[1].  The result is
            # returned as a dict.

            parent_uids = self._apply_filter(etype=etype[0])
            child_uids  = self._apply_filter(etype=etype[1])

            rel = self._description.get('tree', dict())
            for p in parent_uids:

                ret[p] = list()
                if p not in rel:
                    print(('inconsistent : no relations for %s' % p))
                    continue

                for c in rel[p]['children']:
                    if c in child_uids:
                        ret[p].append(c)

        return ret


    # --------------------------------------------------------------------------
    #
    def ranges(self, state=None, event=None, time=None, collapse=True):
        '''
        Gets a set of initial and final conditions, and computes time ranges in
        accordance to those conditions from all session entities. The resulting
        set of ranges is then collapsed to the minimal equivalent set of ranges
        covering the same set of times.

        Please refer to the :class:`Entity.ranges` documentation on detail on
        the constrain parameters.

        Setting 'collapse' to 'True' (default) will prompt the method to
        collapse the resulting set of ranges.
        '''

        ranges = list()
        for uid,entity in list(self._entities.items()):
            try:
                ranges += entity.ranges(state, event, time, collapse=False)
            except ValueError:
                print(('no ranges for %s' % uid))
                # ignore entities for which the conditions did not apply
                # pass

        if not ranges:
            return []

        if collapse:
            ret = ru.collapse_ranges(ranges)
        else:
            ret = ranges

        # sort ranges by start time and return
        return sorted(ret, key=lambda r: r[1])


    # --------------------------------------------------------------------------
    #
    def timestamps(self, state=None, event=None, time=None, first=False):
        '''
        This method accepts a set of conditions, and returns the list of
        timestamps for which those conditions applied, i.e., for which state
        transitions or events are known which match the given 'state' or 'event'
        parameter.  If no match is found, an empty list is returned.

        Both `state` and `event` can be lists, in which case the union of all
        timestamps are returned.

        The `time` parameter is expected to be a single tuple, or a list of
        tuples, each defining a pair of start and end time which are used to
        constrain the matching timestamps.

        If `first` is set to `True`, only the timestamps for the first matching
        events (per entity) are returned.

        The returned list will be sorted.
        '''

        ret = list()
        for _,entity in list(self._entities.items()):
            tmp = entity.timestamps(state=state, event=event, time=time)
            if tmp and first:
                ret.append(tmp[0])
            else:
                ret += tmp

        return sorted(ret)


    # --------------------------------------------------------------------------
    #
    def duration(self, state=None, event=None, time=None, ranges=None):
        '''
        This method accepts the same set of parameters as the `ranges()` method,
        and will use the `ranges()` method to obtain a set of ranges.  It will
        return the sum of the durations for all resulting & collapsed ranges.

        Example::

           session.duration(state=[rp.NEW, rp.FINAL]))

        where `rp.FINAL` is a list of final task states.
        '''

        if not ranges:
            ranges = self.ranges(state, event, time)

        else:
            assert not state
            assert not event
            assert not time

            # make sure the ranges are collapsed (although they likely are
            # already...)
            ranges = ru.collapse_ranges(ranges)

        return sum(r[1] - r[0] for r in ranges)


    # --------------------------------------------------------------------------
    #
    def concurrency(self, state=None, event=None, time=None, sampling=None):
        '''
        This method accepts the same set of parameters as the `ranges()` method,
        and will use the `ranges()` method to obtain a set of ranges.  It will
        return a time series, counting the number of tasks which are
        concurrently matching the ranges filter at any point in time.

        The additional parameter `sampling` determines the exact points in time
        for which the concurrency is computed, and thus determines the sampling
        rate for the returned time series.  If not specified, the time series
        will contain all points at which the concurrency changed.  If specified,
        it is interpreted as second (float) interval at which, after the
        starting point (begin of first event matching the filters) the
        concurrency is computed.

        Returned is an ordered list of tuples::

            [ [time_0, concurrency_0],
              [time_1, concurrency_1],
              ...
              [time_n, concurrency_n] ]

        where `time_n` is represented as `float`, and `concurrency_n` as `int`.

        Example::

            session.filter(etype='task').concurrency(state=[rp.AGENT_EXECUTING,
                rp.AGENT_STAGING_OUTPUT_PENDING])

        '''

        INC =  1  # increase concurrency
        DEC = -1  # decrease concurrency

        ranges = list()
        for _,e in list(self._entities.items()):
            ranges += e.ranges(state, event, time)

        if not ranges:
            return []

        times = list()
        # get all start and end times for all ranges.  The start of a range will
        # increase concurrency by one at that time stamp, and the end will
        # decrease it by one again.
        for r in ranges:
            times.append([r[0], INC])
            times.append([r[1], DEC])

        # sort those times
        times.sort()

        # get the overall concurrency data
        conc = 0
        data = list()  # [[time, concurrency], ...]
        for time, val in times:
            conc += val
            data.append([time, conc])

        # collapse time stamps (use last value on same time stamps)
        collapsed = list()
        last      = data[0]
        for time, val in data[1:]:
            if time != last[0]:
                collapsed.append(last)
            last = [time, val]

        # append last time
        collapsed.append(last)

        # make sure we start at zero (at time of first event)
        collapsed.insert(0, [collapsed[0][0], 0])

        if not sampling:
            # return as is
            ret = collapsed

        else:
            # select data points according to sampling
            # get min time, and create timestamps at regular intervals
            t   = times[0][0]
            v   = collapsed[0][1]
            ret = list()
            for time, val in collapsed:
                while time >= t:
                    ret.append([t, v])
                    t += sampling
                # value is changed after reaching the next timestamp
                v = val

            # append last time stamp if it is not appended, yet
            if ret[-1] != [t, v]:
                ret.append([t, v])

        return ret


    # --------------------------------------------------------------------------
    #
    def rate(self, state=None, event=None, time=None, sampling=None,
            first=False):
        '''
        This method accepts the same parameters as the `timestamps()` method: it
        will count all matching events and state transitions as given, and will
        return a time series of the rate of how many of those events and/or
        transitions occurred per second.

        The additional parameter `sampling` determines the exact points in time
        for which the rate is computed, and thus determines the sampling rate
        for the returned time series.  If not specified, the time series will
        contain all points at which and event occurred, and the rate value will
        only be determined by the time passed between two consecutive events.
        If specified, it is interpreted as second (float) interval at which,
        after the starting point (begin of first event matching the filters) the
        rate is computed.

        Returned is an ordered list of tuples::

          [ [time_0, rate_0] ,
            [time_1, rate_1] ,
            ...
            [time_n, rate_n] ]

        where `time_n` is represented as `float`, and `rate_n` as `int`.

        The `time` parameter is expected to be a single tuple, or a list of
        tuples, each defining a pair of start and end time which are used to
        constrain the resulting time series.

        The 'first' is defined, only the first matching event fir the selected
        entities is considered viable.

        Example::

           session.filter(etype='task').rate(state=[rp.AGENT_EXECUTING])
        '''

        timestamps = self.timestamps(event=event, state=state, time=time,
                                     first=first)

        if not timestamps:
            # nothing to do
            return []


        times = list()
        if sampling:
            # get min and max timestamp, and create sampling points at
            # regular intervals
            r_min = timestamps[0]
            r_max = timestamps[-1]

            t = r_min
            while t < r_max:
                times.append(t)
                t += sampling
            times.append(r_max)

        else:
            # we create an entry at all timestamps
            times = timestamps[:]

        if not times:
            # nothing to do
            return []

        # we need to make sure that no two consecutive timestamps are the same,
        # as that would lead to a division by zero later on
        times = sorted(set(times))

      # import pprint
      # pprint.pprint(times)

        # make sure we start in correct state, and first data point does not
        # occur before sampling starts
        assert timestamps[0] >= times[0]

        # we have the time sequence, now compute event rate at those points
        ts_idx  = 0                # index into the list of timestamps
        ts_len  = len(timestamps)  # number of timestamps
        ret     = list()           # our rate time series
        t_start = times[0]         # current samplint window
        t_stop  = times[0]

        for t in times[1:]:
            t_start = t_stop       # slide sampling window to next sample time
            t_stop  = t
            cnt     = 0            # reset event counter


            while ts_idx < ts_len:
                if timestamps[ts_idx] <= t_stop:
                    # timestamp is in range, count event
                    cnt += 1
                else:
                    # we need to slide the sampling window
                    break
                # go to next timestamp
                ts_idx += 1

          # print 'window: %f - %f (%f) : %5d' % \
          #         (t_start, t_stop, t_stop - t_start, cnt)

            # sampling window completed - store rate of events in sample
            ret.append([t_stop, cnt / (t_stop - t_start)])

        return ret


    # --------------------------------------------------------------------------
    #
    def utilization(self, metrics, rtype='cpu', udurations=None):

        if self._stype != 'radical.pilot':
            raise ValueError('session utilization is only available on '
                             'radical.pilot sessions')

        import radical.pilot as rp

        provided  = rp.utils.get_provided_resources(self, rtype)
        consumed  = rp.utils.get_consumed_resources(self, rtype, udurations)
        stats_abs = {'total':   0.0}
        stats_rel = {'total': 100.0}
        total     = 0.0

        for pid in provided['total']:
            for box in provided['total'][pid]:
                stats_abs['total'] += (box[1] - box[0]) * \
                                      (box[3] - box[2]  + 1)
        total = (max(stats_abs['total'], 1))

        for metric in metrics:
            if isinstance(metric, list):
                name  = metric[0]
                parts = metric[1]
            else:
                name  = metric
                parts = [metric]

            if name not in stats_abs:
                stats_abs[name] = 0.0

            for part in parts:
                for uid in consumed[part]:
                    for box in consumed[part][uid]:
                        stats_abs[name] += (box[1] - box[0]) * \
                                           (box[3] - box[2]  + 1)

        info  = ''
        info += '%s [%d]\n' % (self.uid, len(self.get(etype='task')))

        for metric in metrics + ['total']:
            if isinstance(metric, list):
                name  = metric[0]
                parts = metric[1]
            else:
                name  = metric
                parts = ''

            val = stats_abs[name]
            if val == 0.0: glyph = '!'
            else         : glyph = ''
            rel = 100.0 * val / total
            stats_rel[name] = rel
            info += '    %-20s: %14.3f  %8.3f%%  %2s  %s\n' \
                  % (name, val, rel, glyph, parts)

        have = 0.0
        over = 0.0
        work = 0.0
        for metric in sorted(stats_abs.keys()):
            if metric == 'total':
                have  += stats_abs[metric]
            else:
                if metric == 'Execution Cmd':
                    work  += stats_abs[metric]
                else:
                    over  += stats_abs[metric]

        miss = have - over - work

        rel_over = 100.0 * over / total
        rel_work = 100.0 * work / total
        rel_miss = 100.0 * miss / total

        stats_abs['Other'] = miss
        stats_rel['Other'] = rel_miss

        info += '\n'
        info += '    %-20s: %14.3f  %8.3f%%\n' % ('total', have, 100.0)
        info += '    %-20s: %14.3f  %8.3f%%\n' % ('over',  over, rel_over)
        info += '    %-20s: %14.3f  %8.3f%%\n' % ('work',  work, rel_work)
        info += '    %-20s: %14.3f  %8.3f%%\n' % ('miss',  miss, rel_miss)

        return provided, consumed, stats_abs, stats_rel, info


    # --------------------------------------------------------------------------
    #
    # def utilization_bak(self, owner, consumer, resource, owner_events=None,
    #                                                   consumer_events=None):
    #     '''
    #     Parms:
    #         owner          : The entity name of the owner of the resources
    #         consumer       : The ename of the entity that consumes the resources
    #                         owned by owner
    #         resource       : The type of resources whose utilization is requested,
    #                         eg. Cores, Memory, GPUS etc.
    #         owner_events   : A list of owner's/owners' events that will be used as
    #                         starting and ending points for the resource ownership.
    #         consumer_events: A list of owner's/owners' events that will be used as
    #                         starting and ending points for resource consumption.

    #     Based on these parameters the resources of the owners are collected, as
    #     well as, the times when the consumer(s) used those resources.

    #     Returned is a dictionary of the form::

    #         { 'owner_0': {'range'      : owner_range,
    #                       'resources'  : resource_size,
    #                       'utilization': [[time_0, resource_utilization_0],
    #                                       [time_1, resource_utilization_1],
    #                                       ...
    #                                       [time_n, resource_utilization_n]]},
    #           'owner_1': {'range'      : owner_range,
    #                       'resources'  : resource_size,
    #                       'utilization': [[time_0, resource_utilization_0],
    #                                       [time_1, resource_utilization_1],
    #                                       ...
    #                                       [time_n, resource_utilization_n]]},
    #           ...
    #           'owner_n': {'range'      : owner_range,
    #                       'resources'  : resource_size,
    #                       'utilization': [[time_0, resource_utilization_0],
    #                                       [time_1, resource_utilization_1],
    #                                       ...
    #                                       [time_n, resource_utilization_n]]}

    #     where `time_n` is represented as `float`, `resource_utilization_n` as
    #     `int`, and resource_size is the total resources the owner has.


    #     Example::

    #         s.utilization(owner          = 'pilot',
    #                       consumer       = 'task',
    #                       resource       = 'cores',
    #                       owner_events   = [{ru.EVENT: 'bootstrap_0_start'},
    #                                         {ru.EVENT: 'bootstrap_0_stop' }])
    #                       consumer_events= [{ru.EVENT: 'exec_start'},
    #                                         {ru.EVENT: 'exec_stop' }])
    #     '''
    #     ret = dict()

    #     # Filter the session to get a session of the owners. If that is empty
    #     # return an empty dict

    #     relations = self .describe('relations', [owner, consumer])
    #     if not relations:
    #         return dict()

    #     owners = self.filter(etype=owner, inplace=False)
    #     if not owners:
    #         return dict()

    #     # Filter the self to get the consumers. If none are found, return an
    #     # empty dict.
    #     #
    #     # FIXME: this should return an dict with zero utilization over the full
    #     #        time range the resource exist.
    #     #
    #     for o in owners.get():
    #         owner_id        = o.uid
    #         owner_resources = o.description.get(resource)
    #         owner_range     = o.ranges(event=owner_events)

    #         consumers = self.filter(etype=consumer, uid=relations[owner_id],
    #                                 inplace=False)
    #         if not consumers:
    #             util = [0]

    #         else:
    #             # Go through the consumer entities and create two dictionaries.
    #             # The first keeps track of how many resources each consumer
    #             # consumes, and the second has the ranges based on the events.
    #             consumer_resources = dict()
    #             consumer_ranges    = dict()

    #             for c in consumers.get():

    #                 ranges  = c.ranges(event=consumer_events)
    #                 cons_id = c.uid

    #                 resources_acquired = 0
    #                 if resource == 'cores':
    #                     cores   = c.description['cpu_processes'] * \
    #                               c.description['cpu_threads']
    #                     resources_acquired += cores
    #                 elif resource == 'gpus':
    #                     gpus    = c.description['gpu_processes']
    #                     resources_acquired += len(gpus)
    #                 else:
    #                     raise ValueError('unsupported utilization resource')

    #                 consumer_resources[cons_id] = resources_acquired

    #                 # Update consumer_ranges if there is at least one range
    #                 if ranges:
    #                     consumer_ranges.update({cons_id: ranges})


    #             # Sort consumer_ranges based on their values. This command
    #             # returns a dictionary, which is sorted based on the first value
    #             # of each entry. In the end the key, are out of order but the
    #             # values are.
    #             consumer_ranges = sorted(iter(list(consumer_ranges.items())),
    #                                      key=lambda k_v: (k_v[1][0],k_v[0]))

    #             # Create a timeseries that contains all moments in consumer
    #             # ranges and sort. This way we have a list that has time any
    #             # change has happened.
    #             times = list()
    #             for cons_id,ranges in consumer_ranges:
    #                 for r in ranges:
    #                     times.append(r[0])
    #                     times.append(r[1])
    #             times.sort()

    #             # we have the time sequence, now compute utilization
    #             # at those points
    #             util = list()
    #             for t in times:
    #                 cnt = 0
    #                 for cons_id,ranges in consumer_ranges:
    #                     for r in ranges:
    #                         if t >= r[0] and t <= r[1]:
    #                             cnt += consumer_resources[cons_id]

    #                 util.append([t, cnt])

    #         ret[owner_id] = {'range'      : owner_range,
    #                          'resources'  : owner_resources,
    #                          'utilization': util}
    #     return ret


    # --------------------------------------------------------------------------
    #
    def consistency(self, mode=None):
        '''
        Performs a number of data consistency checks, and returns a set of UIDs
        for entities which have been found to be inconsistent. The method
        accepts a single parameter `mode` which can be a list of strings
        defining what consistency checks are to be performed. Valid strings are:

        - state_model: check if all entity states are in adherence to the
          respective entity state model
        - event_model: check if all entity events are in adherence to the
          respective entity event model
        - timestamps: check if events and states are recorded with correct
          ordering in time.

        If not specified, the method will execute all three checks.

        After this method has been run, each checked entity will have more
        detailed consistency information available via::

            entity.consistency['state_model'] (bool)
            entity.consistency['event_model'] (bool)
            entity.consistency['timestamps' ] (bool)
            entity.consistency['log' ]        (list of strings)

        The boolean values each indicate consistency of the respective test, the
        `log` will contain human readable information about specific consistency
        violations.
        '''

        # FIXME: we could move the method to the entity, so that we can check
        #        consistency for each entity individually.

        self._rep.header('running consistency checks')

        ret   = list()
        MODES = ['state_model', 'event_model', 'timestamps']

        if not mode:
            mode = MODES

        if not isinstance(mode, list):
            mode = [mode]

        for m in mode:
            if m not in MODES:
                raise ValueError('unknown consistency mode %s' % m)

        if 'state_model' in mode:
            ret.extend(self._consistency_state_model())

        return list(set(ret))  # make list unique


    # --------------------------------------------------------------------------
    #
    def usage(self, alloc_entity, alloc_events,
                    block_entity, block_events,
                    use_entity,   use_events):
        '''
        This method creates a dict with three entries: `alloc`, `block`, `use`.
        Those three dict entries in turn have a a dict of entity IDs for all
        entities which have blocks in the respective category, and foreach of
        those entity IDs the dict values will be a list of rectangles.

        A resource is considered:

        - `alloc` (allocated) when it is owned by the RCT application;
        - `block` (blocked) when it is reserveed for a specific task;
        - `use` (used) when it is utilized by that task.

        Each of the rectangles represents a continuous block of resources which
        is alloced/blocked/used:

        - x_0 time when `alloc/block/usage` begins;
        - x_1 time when `alloc/block/usage` ends;
        - y_0 lowest index of a continuous block of resource IDs;
        - y_1 highest index of a continuous block of resource IDs.

        Any specific entity (pilot, task) can have a **set** of such resource
        blocks, for example, a task might be placed over multiple,
        non-consecutive nodes:

        - gpu and cpu resources are rendered as separate blocks (rectangles).

        Args:
            alloc_entity (Entity): :class:`Entity` instance which allocates
                resources
            alloc_events (list): event tuples which specify allocation time
            block_entity (Entity): :class:`Entity` instance which blocks
                resources
            block_events (list): event tuples which specify blocking time
            use_entity (Entity): :class:`Entity` instance which uses resources
            use_events (list): event tuples which specify usage time

        Example::

            usage('pilot', [{ru.STATE: None, ru.EVENT: 'bootstrap_0_start'},
                            {ru.STATE: None, ru.EVENT: 'bootstrap_0_stop' }],
                  'task' , [{ru.STATE: None, ru.EVENT: 'schedule_ok'      },
                            {ru.STATE: None, ru.EVENT: 'unschedule_stop'  }],
                  'task' , [{ru.STATE: None, ru.EVENT: 'exec_start'       },
                            {ru.STATE: None, ru.EVENT: 'exec_stop'        }])
        '''

        # this is currently only supported for RP sessions, as we only know for
        # pilots and tasks how to dig resource information out of session and
        # entity metadata.
        assert self.stype == 'radical.pilot', \
               'stype %s unsupported' % self._stype

        # for RP sessions, create resource indices which can be used to
        # determine the y-axis values for the rectangles.  This is basically
        # a dict of node_names for each alloc_entity which points to two indexes
        # for each node: one starting index for GPUs, and one for CPU cores
        res_idx = dict()

        idx = 0
        for ae in self.get(etype=alloc_entity):
            uid   = ae.uid
            nodes = ae.cfg['resource_details']['rm_info']['node_list']
            cpn   = ae.cfg['resource_details']['rm_info']['cores_per_node']
            gpn   = ae.cfg['resource_details']['rm_info']['gpus_per_node']

            if ae.uid not in res_idx:
                res_idx[ae.uid] = dict()

            res_idx[ae.uid]['_min'] = idx

            for n in nodes:
                res_idx[ae.uid][n[1]] = [[idx      , idx + cpn       - 1],
                                         [idx + cpn, idx + cpn + gpn - 1]]
                idx += cpn
                idx += gpn

            res_idx[ae.uid]['_max'] = idx - 1

        # ----------------------------------------------------------------------
        # RP specific helper method to convert given entity information into
        # a set of y-value ranges.  This returns a tuple of lists where each
        # list contains tuples of resource indexes (y-values)
        def get_res_idx(entity):

            if entity.etype == 'pilot':

              # print '----'
              # print entity.uid
              # import pprint
              # pprint.pprint(res_idx[entity.uid])

                # we assume min/max to cover CPU and GPU
                return [[[res_idx[entity.uid]['_min'],
                          res_idx[entity.uid]['_max']]], []]

            elif entity.etype == 'task':

                # find owning pilot
                pid = entity.cfg.get('pilot')
                if not pid:
                    # no resources used
                    return [[],[]]

                cpu_idx = list()
                gpu_idx = list()
                for slot in entity.cfg['slots']['nodes']:
                    node_id = slot['uid']
                    for cslot in slot['core_map']:
                        for c in cslot:
                            cpu_idx.append(res_idx[pid][node_id][0][0] + c)
                    for gslot in slot['gpu_map']:
                        for g in gslot:
                            gpu_idx.append(res_idx[pid][node_id][1][0] + g)

                # identify continuous groups of y-values and return
                return [
                        [list(g) for g in mit.consecutive_groups(cpu_idx)],
                        [list(g) for g in mit.consecutive_groups(gpu_idx)]
                       ]
        # ----------------------------------------------------------------------

        ret = {'alloc': dict(),
               'block': dict(),
               'use'  : dict()}

        etypes = {'alloc': {'etype' : alloc_entity,
                            'events': alloc_events},
                  'block': {'etype' : block_entity,
                            'events': block_events},
                  'use'  : {'etype' : use_entity,
                            'events': use_events}}

        for entity in self.get():

            for mode in ['alloc', 'block', 'use']:

                if etypes[mode]['etype'] == entity.etype:

                    event_list = etypes[mode]['events']
                    if not event_list:
                        continue

                    if not isinstance(event_list[0], list):
                        event_list = [event_list]

                    uid = entity.uid
                    if uid not in ret[mode]:
                        ret[mode][uid] = list()

                    for events in event_list:
                        assert len(events) == 2

                        y_values = get_res_idx(entity)
                        t_values = entity.ranges(event=events)

                      # print
                      # print uid
                      # print events
                      #
                      # import pprint
                      # pprint.pprint(t_values)
                      # pprint.pprint(y_values)

                        for t_range in t_values:
                            for y_range in y_values[0]:
                                ret[mode][uid].append([t_range[0], t_range[-1],
                                                       y_range[0], y_range[-1]])
                            for y_range in y_values[1]:
                                ret[mode][uid].append([t_range[0], t_range[-1],
                                                       y_range[0], y_range[-1]])

        return ret


    # --------------------------------------------------------------------------
    #
    def _consistency_state_model(self):

        ret = list()  # list of inconsistent entity IDs

        for et in self.list('etype'):

            self._rep.info('%s state model\n' % et)
            # sm = self.describe('state_model', etype=et)
            sv = self.describe('state_values', etype=et)[et]['state_values']

            for e in self.get(etype=et):

                es = e.states

                if not sv:
                    if es:
                        self._rep.warn('  %-30s : %s' % (et, list(es.keys())))
                        e._consistency['state_model'] = None
                    continue

                self._rep.info('  %-30s :' % e.uid)

                missing = False   # did we miss any state so far?
                final_v = sorted(sv.keys())[-1]
                final_s = sv[final_v]

                if not isinstance(final_s, list):
                    final_s = [final_s]

                sm_ok    = True
                sm_log   = list()
                miss_log = list()
                for s in sv.values():

                    if not s:
                        continue

                    if not isinstance(s, list):
                        s = [s]

                    # check if we have that state
                    found = None
                    for _s in s:
                        if _s in es:
                            found = _s
                            break

                    if found:

                        if missing:

                            if found not in final_s:
                                # found a state after a previous one was missing,
                                # but we are not final.  Oops
                                self._rep.warn('+')
                                sm_log.extend(miss_log)
                                miss_log = list()
                                sm_ok = False
                                continue

                    else:
                        if s == final_s:
                            # no final state?  Oops
                            self._rep.error('no final state! ')
                            sm_ok = False
                            sm_log.append('missing final state')
                            continue

                        else:
                            # Hmm, might be ok.  Lets see...
                            missing = True
                            self._rep.warn('*')
                            miss_log.append('missing state(s) %s' % s)
                            continue

                    self._rep.ok('+')

                e._consistency['state_model'] = sm_ok
                e._consistency['log'].extend(sm_log)

                if not sm_ok:
                    ret.append(e.uid)

                self._rep.plain('\n')

        return ret


    # --------------------------------------------------------------------------
    #
    def tzero(self, t):
        '''
        Setting a `tzero` timestamp will shift all timestamps for all entities
        in this session by that amount.  This simplifies the alignment of
        multiple sessions, or the focus on specific events.
        '''

        old_tzero   = self._tzero
        self._tzero = t

        for entity in list(self._entities.values()):

            # entity.states are shallow copies of the events
            for event in entity.events:
                event[ru.TIME] += old_tzero
                event[ru.TIME] -= self._tzero


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import sys

import matplotlib        as mpl
import matplotlib.pyplot as plt

import radical.analytics as ra

from radical.analytics.utils import to_latex

# This utilization plot accounts for resource usages the following way:
#
#   - initially, all resources are owned by the 'system'.
#   - if a pilot gets placed (`bootstrap_0_start` event), the resources' control
#     transitions from `system` to `bootstrap`.
#   - if bootstrapper is done (at `PMGR_ACTIVE` state of the pilot), they
#     transition from `bootstrap` to `idle`
#   - etc etc.
#
# Basicall, specific profile events signal that control over a certain amount of
# resources transitions from one entity to another.  Another example
#
#   - the task event `schedule_ok` signals that some resources transition out of
#     the idle pool and are now allocated for a task.  The task may not yet be
#     running, so the resources transition to `rp_exec`, signifying that RP
#     still works to execute that task.
#   - `unschedule_stop` signals that execution and cleanup completed, that
#     the resources are not allocated for a task anymore and are returned back
#     into the `idle` pool.
#
# The plotting content is controled as follows:
#
#   - the `metrics` data strcuture lists all metrics to be plotted, i.e., all
#     'resource owners' in the sense above (e.g., `bootstrap`, `idle`, `rp_exec`
#     etc.
#
#     metrics  = [  # metric,      line color, alpha, fill color, alpha
#                   ['bootstrap', ['#c6dbef',  0.0,   '#c6dbef',  1  ]],
#                   [...]
#                ]
#
#   - the `tmap` data structure defines what entities' events will be considered
#     to determine resource transitions:
#
#     tmap = {
#             # etype  : transition map
#             'pilot'  : p_trans,
#             ...
#            }
#
#   - the individual `tmap` entries (transition maps) determine what *events*
#     signal transition *from* what entity *to* what other entity:
#
#     t_trans = [
#               # event          , from,       to
#             [{1: 'schedule_ok'}, 'idle'    , 'exec_rp' ],
#             [{1: 'exec_start'} , 'exec_rp' , 'exec_launch' ],
#             [{1: 'rank_start'} , 'exec_launch' , 'exec_task'],
#
#
# ----------------------------------------------------------------------------
#
plt.style.use(ra.get_mplstyle("radical_mpl"))



# ------------------------------------------------------------------------------
# pick and choose what resources to plot (one sub-plot per resource)
resrc = ['cpu', 'gpu']
resrc = ['cpu']

# pick and choose what contributions to plot
metrics = [  # metric        , [color    , alpha]]
             ['bootstrap'    , ['#00FF00',  0.4 ]],
             ['agent'        , ['#00FF00',  0.6 ]],
             ['raptor_master', ['#FF0000',  0.2 ]],
             ['raptor_task'  , ['#FF0000',  1.0 ]],
             ['raptor_worker', ['#FF0000',  0.4 ]],
             ['exec_task'    , ['#0000FF',  1.0 ]],
             ['exec_launch'  , ['#0000FF',  0.4 ]],
             ['exec_rp'      , ['#0000FF',  0.2 ]],
           # ['term'         , ['#00FF00',  0.3 ]],
             ['idle'         , ['#00FF00',  0.1 ]]
          ]

labels = [['exec_rp',       'exec_launch',   'exec_task',   'bootstrap', 'agent'],
          ['raptor_master', 'raptor_worker', 'raptor_task', 'idle']]

# ------------------------------------------------------------------------------
# transition events for pilot, task, master, worker, request
#
# event  : resource transitions from : resource transitions to
#
p_trans = [
        [{1: 'bootstrap_0_start'} , 'system'        , 'bootstrap'    ] ,
        [{5: 'PMGR_ACTIVE'}       , 'bootstrap'     , 'idle'         ] ,
        [{6: 'cancel_pilot'}      , 'idle'          , 'term'         ] ,
        [{1: 'bootstrap_0_stop'}  , 'term'          , 'system'       ] ,
        [{1: 'sub_agent_start'}   , 'idle'          , 'agent'        ] ,
        [{1: 'sub_agent_stop'}    , 'agent'         , 'term'         ]
]

t_trans = [
        [{1: 'schedule_ok'}       , 'idle'          , 'exec_rp'      ] ,
        [{1: 'exec_start'}        , 'exec_rp'       , 'exec_launch'  ] ,
        [{1: 'rank_start'}        , 'exec_launch'   , 'exec_task'    ] ,
     #  [{1: 'app_start'}         , 'exec_task'     , 'exec_task'    ] ,
     #  [{1: 'app_stop'}          , 'exec_task'     , 'exec_task'    ] ,
        [{1: 'rank_stop'}         , 'exec_task'     , 'exec_launch'  ] ,
        [{1: 'exec_stop'}         , 'exec_launch'   , 'exec_rp'      ] ,
        [{1: 'unschedule_stop'}   , 'exec_rp'       , 'idle'         ]
]

m_trans = [
        [{1: 'schedule_ok'}       , 'idle'          , 'exec_rp'      ] ,
        [{1: 'exec_start'}        , 'exec_rp'       , 'exec_launch'  ] ,
        [{1: 'rank_start'}        , 'exec_launch'   , 'raptor_master'] ,
        [{1: 'rank_stop'}         , 'raptor_master' , 'exec_launch'  ] ,
        [{1: 'exec_stop'}         , 'exec_launch'   , 'exec_rp'      ] ,
        [{1: 'unschedule_stop'}   , 'exec_rp'       , 'idle'         ]
]

w_trans = [
        [{1: 'schedule_ok'}       , 'idle'          , 'exec_rp'      ] ,
        [{1: 'exec_start'}        , 'exec_rp'       , 'exec_launch'  ] ,
        [{1: 'rank_start'}        , 'exec_launch'   , 'raptor_worker'] ,
        [{1: 'rank_stop'}         , 'raptor_worker' , 'exec_launch'  ] ,
        [{1: 'exec_stop'}         , 'exec_launch'   , 'exec_rp'      ] ,
        [{1: 'unschedule_stop'}   , 'exec_rp'       , 'idle'         ]
]

r_trans = [
        [{1: 'rank_start'}        , 'raptor_worker' , 'raptor_task'  ] ,
        [{1: 'rank_stop'}         , 'raptor_task'   , 'raptor_worker']
]

# what entity maps to what transition table
tmap = {
        'pilot'         : p_trans,
        'task'          : t_trans,
        'raptor.master' : m_trans,
        'raptor.worker' : w_trans,
        'raptor.task'   : r_trans,
}


# ==============================================================================

# metrics to stack and to plot
to_stack = [m[0]       for m in metrics]
to_plot  = {m[0]: m[1] for m in metrics}

# Use to set Y-axes to % of resource utilization
use_percent = True


# ------------------------------------------------------------------------------
#
def main():

    # read the session profiles
    src  = sys.argv[1]

    if len(sys.argv) == 3: stype = sys.argv[2]
    else                 : stype = 'radical.pilot'

    session = ra.Session.create(src=src, stype=stype)

    # this script only works for one pilot
    pilots = session.get(etype='pilot')
    assert len(pilots) == 1, len(pilots)

    sid     = session.uid
    pilot   = pilots[0]
    rm_info = pilot.cfg['resource_details']['rm_info']
    p_size  = pilot.description['cores']
    n_nodes = int(p_size / rm_info['cores_per_node'])
    n_tasks = len(session.get(etype='task'))

    # Derive pilot and task timeseries of a session for each metric
    p_resrc, series, x = ra.get_pilot_series(session, pilot, tmap, resrc, use_percent)

    # #plots = # of resource types (e.g., CPU/GPU = 2 resource types = 2 plots)
    n_plots = 0
    for r in p_resrc:
        if p_resrc[r]:
            n_plots += 1

    # sub-plots for each resource type, legend on first, x-axis shared
    fig = plt.figure(figsize=(ra.get_plotsize(400)))
    gs  = mpl.gridspec.GridSpec(n_plots, 1)

    for plot_id, r in enumerate(resrc):

        if not p_resrc[r]:
            continue

        # create sub-plot
        ax = plt.subplot(gs[plot_id])

        # stack timeseries for each metrics into areas
        areas = ra.stack_transitions(series, r, to_stack)

        # plot individual metrics
        prev_m  = None
        patches = dict()
        legend  = dict()
        for m in areas:

            if m not in to_plot:
                if m != 'time':
                    print('skip', m)
                continue

            pcol   = to_plot[m][0]
            palpha = to_plot[m][1]

            # plot the (stacked) areas
            # fill first metric toward 0, all others towards previous line
            if not prev_m:
                patch = ax.fill_between(areas['time'], areas[m],
                                        step='post', label=m, linewidth=0.0,
                                        color=pcol, alpha=palpha)

            else:
                patch = ax.fill_between(areas['time'], areas[m], areas[prev_m],
                                        step='post', label=m, linewidth=0.0,
                                        color=pcol, alpha=palpha)

            # remember patches for legend
            legend[m] = to_latex(m)
            patches[m] = patch

            # remember this line to fill against
            prev_m = m

        ax.set_xlim([x['min'], x['max']])
        if use_percent:
            ax.set_ylim([0, 110])
        else:
            ax.set_ylim([0, p_resrc[r]])

        ax.set_xlabel(to_latex('time (s)'))
        ax.set_ylabel(to_latex('%s (%%)' % r))

        # resort the legend to list entries from left to right
        ncol = len(labels[0])
        label_order = list()
        for i in range(ncol):
            for row in labels:
                if len(row) > i:
                    label_order.append(row[i])
        legend_sorted  = list()
        patches_sorted = list()
        for label in label_order:
            legend_sorted.append(legend[label])
            patches_sorted.append(patches[label])

        # first sub-plot gets legend
        if plot_id == 0:
            ax.legend(patches_sorted, legend_sorted, loc='upper center',
                    ncol=ncol, bbox_to_anchor=(0.5, 1.25), fancybox=True,
                    shadow=True)

    for ax in fig.get_axes():
        ax.label_outer()

    # Title of the plot
    fig.suptitle(to_latex('%s Tasks - %s Nodes' % (n_tasks, n_nodes)))

    # Save a publication quality plot
    fig.savefig('%s_util2.png' % sid, dpi=600, bbox_inches='tight')


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    main()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import sys
import pprint
import radical.analytics as ra

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


"""
This example illustrates the use of the method ra.Session.describe()
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src     = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # and here we go. Session.describe() can be used to print the entities state
    # models, the entities runtime event models, and the state values. The state
    # models of each entity of our session can be described with:
    ppheader("state models of all the entities of the session")
    pprint.pprint(session.describe('state_model'))

    # The keys of the innermost dictionary are the names of the state, the
    # values their time precedence expressed as an integer. Given two states
    # with different integers, the state with the smaller integer is always
    # guaranteed to happen before the state with the larger integer. The states
    # with equal integers are guaranteed to be mutual exclusive. Currently,
    # RADICAL-Pilots has three states with integer 15. These are the three final
    # states in which each entity can end its life-cycle and, accordingly to the
    # state model, are mutually exclusive.
    #
    # We can restrict the entities for which to print the state model with:
    ppheader("state models of the entities of type 'unit' and 'pilot'")
    pprint.pprint(session.describe('state_model', etype=['unit', 'pilot']))

    # Note that the list can be omitted when passing a single value to etype or
    # any other argument key:
    ppheader("state models of the entities of type 'unit'")
    pprint.pprint(session.describe('state_model', etype='unit'))

    # The ordered sequence of states can be described by using:
    ppheader("Ordered sequence of states")
    pprint.pprint(session.describe('state_values'))

    # We can use similar calls to describe the events of every entity of the
    # session:
    ppheader("runtime event model of all the entities of the session")
    pprint.pprint(session.describe('event_model'))

    # or the relations among all the entities of the session:
    ppheader("relations among all the entities of the session")
    pprint.pprint(session.describe('relations', etype=['pilot', 'unit']))

    # We can restrict the type of entities to describe also for the last two
    # calls:
    ppheader("runtime event model for the entities of type 'unit'")
    pprint.pprint(session.describe('event_model', etype='unit'))

    sys.exit(0)



#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys
import pprint

import radical.utils as ru
import radical.pilot as rp
import radical.analytics as ra

"""
This example illustrates hoq to obtain durations for arbitrary (non-state)
profile events.
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # First we look at the *event* model of our session.  The event model is
    # usually less stringent than the state model: not all events will always be
    # available, events may have certain fields missing, they may be recorded
    # multiple times, their meaning may slightly differ, depending on the taken
    # code path.  But in general, these are the events available, and theeir
    # relative ordering.
    ppheader("event models")
    pprint.pprint(session.describe('event_model'))
    pprint.pprint(session.describe('statistics'))

    # Let's say that we want to see how long RP took to prepare the execution of
    # units.  That time is from when the unit reached the `AGENT_EXECUTING`
    # state (as it now enters the executing component), up to when RP passes
    # control over unit execution to the system layer.  The latter is signalled
    # by the event named `exec_start`.
    #
    # In order to have a complete pair of initial and final conditions to time,
    # we express the state condition also as event.
    #
    # We first get the respective time ranges for all units, just to have a look
    # at them, and then compute the overall duration
    ppheader("Time spent by the units in exec-preparation")
    units = session.filter(etype='unit', inplace=False)
    print('#units   : %d' % len(units.get()))

    ranges = units.ranges(event=[{ru.EVENT: 'state',
                                  ru.STATE: rp.AGENT_EXECUTING},
                                 {ru.EVENT: 'exec_start'}],
                          collapse=False)
    print('ranges   :')
    for r in ranges:
        print('  [%7.2f, %7.2f] = %7.2f' % (r[0], r[1], r[1] - r[0]))

    duration = units.duration(ranges=ranges)
    print('duration : %.2f' % duration)


    # now perform a sanity check: for each unit we check if the duration as
    # obtained above is in fact smaller than the duration for the
    # `AGENT_EXECUTING` state, as one would expect.
    ppheader("sanity check: exec time > prep time")
    oopses = list()
    for unit in units.get():

        prep_duration = unit.duration(event=[{ru.EVENT: 'state',
                                              ru.STATE: rp.AGENT_EXECUTING},
                                             {ru.EVENT: 'exec_start'}])
        exec_duration = unit.duration(event=[{ru.EVENT: 'state',
                                              ru.STATE: rp.AGENT_EXECUTING},
                                             {ru.EVENT: 'state',
                                              ru.STATE: rp.AGENT_STAGING_OUTPUT_PENDING}])
        diff = exec_duration - prep_duration
        print('%7.2f > %7.2f : %s' % (exec_duration, prep_duration, diff > 0))

        # we could in principle check against session accuracy in this place,
        # but we do happen to know that both events are recorded in the same
        # component, and time should thus always be linear.
        if diff <= 0:
            oopses.append([unit.uid, diff])

    # now perform a sanity check: for each unit we check if the duration as
    # obtained above is in fact smaller than the duration for the
    # `AGENT_EXECUTING` state, as one would expect.
    ppheader("pure exec times (exec_start ... exec_stop)")
    durations=list()
    for unit in units.get():
        exec_duration = unit.duration(event=[{ru.EVENT: 'exec_start'},
                                             {ru.EVENT: 'exec_stop'}])
        print('%s: %7.2f' % (unit.uid, exec_duration))
        durations.append(exec_duration)

    print('average    : %7.2f' % (sum(durations) / len(durations)))
    print()

    ppheader("concurrent units in between exec_start and exec_stop events")
    concurrency = units.concurrency(event=[{ru.EVENT: 'exec_start'},
                                           {ru.EVENT: 'exec_stop' }],
                                    sampling=10)
    pprint.pprint(concurrency)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2024, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
# we *first* import radical.utils, so that the monkeypatching of the logger has
# a chance to kick in before the logging module is pulled by any other 3rd party
# module, and also to monkeypatch `os.fork()` for the `atfork` functionality
#
import os            as _os
import radical.utils as _ru


# ------------------------------------------------------------------------------
#
from .experiment import Experiment
from .session    import Session
from .entity     import Entity
from .plotter    import Plotter


from .utils import get_plotsize, get_mplstyle, stack_transitions
from .utils import get_pilot_series, get_plot_utilization, get_pilots_zeros
from .utils import to_latex
from .utils import tabulate_durations


# ------------------------------------------------------------------------------
#
# get version info
#
_mod_root = _os.path.dirname (__file__)

version_short, version_base, version_branch, version_tag, version_detail \
             = _ru.get_version(_mod_root)
version      = version_short
__version__  = version_detail


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys
import pprint

import numpy             as np
import matplotlib        as mpl
import matplotlib.pyplot as plt

import radical.analytics as ra


# This script plots the core utilization of a set of RP sessions in a stacked
# barplot, where the stacked elements represent how the cores were (or were not)
# utilized.  The elements will always add up to 100%, thus the whole bar
# represents the amount of core-hours available to the session.  We only look at
# events on the agent side.
#
# The bar elements are determined as follows (for each pilot)
#
#  - During bootstrap, configuration, setup, and termination, cores are
#    essentially unused.  We separate out the times needed for those steps, and
#    multiply by pilot size, to derive the number of core hours essentially
#    spent on those activities (see `PILOT_DURATIONS` below).
#
#    NOTE: we assume that the pilot activities thus measured stop at the point
#          when the first unit gets scheduled on a (set of) core(s), and
#          restarts when the last unit gets unscheduled, as that is the time
#          frame where we in principle consider cores to be available to the
#          workload.
#
#  - For each unit, we look at the amount of time that unit has been scheduled
#    on a set of cores, as those cores are then essentially blocked.  Multiplied
#    by the size of the unit, that gives a number of core-hours those cores are
#    'used' for that unit.
#
#    not all of that time is utilized for application use though: some is spent
#    on preparation for execution, on spawning, unscheduling etc.  We separate
#    out those utilizations for each unit.
#
#  - we consider core hours to be additive, in the following sense:
#
#    - all core-hours used by the pilot in various global activities listed in
#      the first point, plus the sumof core hours spend by all units in various
#      individual activities as in the second point, equals the overall core
#      hours available to the pilot.
#
#      This only holds with one caveat: after the agent started to work on unit
#      execution, some cores may not *yet* be allocated (scheduler is too slow),
#      or may not be allocated *anymore* (some units finished, we wait for the
#      remaining ones).  We consider those core-hours as 'idle'.
#
#      Also, the agent itself utilizes one node, and we consider that time as
#      agent utilization overhead.
#
# NOTE: we actually use core-seconds instead of core-hours.  So what?!
#

metrics_prte = [

        ['Agent Nodes',       ['agent']],
        ['Pilot Startup',     ['boot', 'setup_1']],
        ['Warmup',            ['warm' ]],
        ['Prepare Execution', ['exec_queue', 'exec_prep']],
        ['Execution RP',      ['exec_rp', 'exec_sh', 'term_sh', 'term_rp']],
        ['Execution PRTE',    ['prte_phase_1', 'prte_phase_2', 'prte_phase_3']],
        ['Execution Cmd',     ['exec_cmd']],
        ['Unschedule',        ['unschedule']],
        ['Draining',          ['drain']],
        ['Pilot Termination', ['term' ]],
        ['Idle',              ['idle' ]],
]


metrics_default = [

        ['Agent Nodes',       ['agent']],
        ['Pilot Startup',     ['boot', 'setup_1']],
        ['Warmup',            ['warm' ]],
        ['Prepare Execution', ['exec_queue', 'exec_prep']],
        ['Pilot Termination', ['term' ]],
        ['Execution RP',      ['exec_rp', 'exec_sh', 'term_sh', 'term_rp']],
        ['Execution Cmd',     ['exec_cmd']],
        ['Unschedule',        ['unschedule']],
        ['Draining',          ['drain']],
        ['Idle',              ['idle' ]],
]

metrics = metrics_default


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    sources = sys.argv[1:]
    xkeys   = dict()  # x-axis labels
    sids    = list()  # need to sort SIDs

    exp = ra.Experiment(sources, stype='radical.pilot')

    # get the numbers we actually want to plot
    for session in exp.sessions:

        sid = session.uid
        n_units = len(session.get(etype='unit'))
        p_size  = 0
        for pilot in session.get(etype='pilot'):
            p_size += pilot.description['cores']

        xkeys[sid] = [n_units, p_size]
        sids.append(sid)

    # sort sessions by pilot size
    sids = sorted(sids, key=lambda sid: xkeys[sid][1])


    # get utilization information
    provided, consumed, stats_abs, stats_rel, info = exp.utilization(metrics=metrics)
  # provided, consumed, stats_abs, stats_rel = exp.utilization(metrics='/path/metrics.json')

    with open('%s.stats' % sid, 'w') as fout:
        fout.write('\n%s\n\n' % info)

  # pprint.pprint(provided)
  # pprint.pprint(consumed)
  # pprint.pprint(stats_abs)
  # pprint.pprint(stats_rel)

    cmap = mpl.cm.get_cmap('tab20c')

    # --------------------------------------------------------------------------
    # core utilization over time (box plot)
    for sid in sids:
        fig  = plt.figure(figsize=(20,14))
        ax   = fig.add_subplot(111)

        step   = 1.0  / (len(metrics) + 1)
        this   = step / 1.0
        legend = list()

        x_min = None
        x_max = None
        y_min = None
        y_max = None

        for metric in metrics:

            color = cmap(this)
            this += step

            legend.append(mpl.lines.Line2D([0], [0], color=color, lw=6))

            if isinstance(metric, list):
                name  = metric[0]
                parts = metric[1]
            else:
                name  = metric
                parts = [metric]

            for part in parts:
                for uid in sorted(consumed[sid][part]):
                    for block in consumed[sid][part][uid]:
                        orig_x = block[0]
                        orig_y = block[2] - 0.5
                        width  = block[1] - block[0]
                        height = block[3] - block[2] + 1.0

                        if x_min is None: x_min = orig_x
                        if x_max is None: x_max = orig_x + width
                        if y_min is None: y_min = orig_x
                        if y_max is None: y_max = orig_x + height

                        x_min = min(x_min, orig_x)
                        y_min = min(y_min, orig_y)
                        x_max = max(x_max, orig_x + width)
                        y_max = max(y_max, orig_y + height)

                        patch = mpl.patches.Rectangle((orig_x, orig_y),
                                                      width, height,
                                                      facecolor=color,
                                                      edgecolor='black',
                                                      fill=True, lw=0.0)
                        ax.add_patch(patch)

        ax.legend(legend, [m[0] for m in metrics], ncol=6,
                   loc='upper center', bbox_to_anchor=(0.5,1.11))
        plt.xlabel('runtime [s]')
        plt.ylabel('resource slot (index)')

        plt.xlim([x_min, x_max])
        plt.ylim([y_min, y_max])
      # plt.xticks(list(range(int(x_min)-1, int(x_max)+1)))
        fig.savefig('%s_core_allocation.png' % sid)
        plt.show()


    # --------------------------------------------------------------------------
    # utilization: contributions as stacked barplot
    #
    # yes, stacked barplot is this cumbersome:
    # http://matplotlib.org/examples/pylab_examples/bar_stacked.html
    #
    plt.figure(figsize=(20,14))
    bottom = np.zeros(len(exp.sessions))
    ind    = np.arange(len(exp.sessions))  # locations of bars on x-axis
    width  = 0.35                          # width of bars

    labels = list()
    plots  = list()

    for metric in metrics + ['Other']:

        color = cmap(this)
        this += step

        legend.append(mpl.lines.Line2D([0], [0], color=color, lw=6))

        if isinstance(metric, list):
            name  = metric[0]
            parts = metric[1]
        else:
            name  = metric
            parts = [metric]

        values = [stats_rel[sid][name] for sid in sids]
        plots.append(plt.bar(ind, values, width, bottom=bottom))
        bottom += values
        labels.append(name)

    if False: plt.ylabel('utilization (% of total resources)')
    else    : plt.ylabel('utilization (in core-seconds)')

    plt.xlabel('#CU / #cores')
    plt.ylabel('utilization (% of total resources)')
    plt.title ('pilot utilization over workload size (#units)')
    plt.xticks(ind, ['%s / %s' % (xkeys[sid][0], xkeys[sid][1]) for sid in sids])

    plt.legend([p[0] for p in plots], labels, ncol=5, loc='upper left',
               bbox_to_anchor=(0,1.13))
    plt.savefig('08c_core_utilization.png')
    plt.show()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import os
import sys
import glob
import pprint
import radical.utils as ru
import radical.pilot as rp
import radical.analytics as ra

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


"""
This example illustrates the use of the method ra.Session.list()
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src     = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # and here we go. Once constructed, our session contain information about
    # the entities of the RADICAL-Cybertool that we are using. We can list these
    # entities and their properties with:
    pnames = session.list()
    ppheader("name of the properties of the session")
    pprint.pprint(pnames)

    # Each entity has at least four properties--etype, uid, state, and
    # event--and we can indipendently list one or more of these properties. The
    # following list the types of every entity in the session:
    etypes = session.list('etype')
    ppheader("name of the types of entity of the session")
    pprint.pprint(etypes)

    # The unique identifier of the entities (note that the identifier is
    # guaranteed to be unique within the scope of the given session. This means
    # that given two session, the same identifier may be used in both of them):
    ppheader("unique identifiers (uid) of all entities of the session")
    uids = session.list('uid')
    pprint.pprint(uids)

    # The name of the states of the entities:
    ppheader("unique names of the states of all entities of the session")
    states = session.list('state')
    pprint.pprint(states)

    # and the name of the events of the entities:
    ppheader("unique names of the events of all entities of the session")
    events = session.list('event')
    pprint.pprint(events)

    # Finally, when useful, we can list subset of properties by using a list
    # notation:
    ppheader("names of the types and states of entity of the session")
    etypes_states = session.list(['etype', 'state'])
    pprint.pprint(etypes_states)

    sys.exit(0)




# coding: utf-8

# In[1]:


import os
import sys
from glob import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pprint import pprint
import json

import radical.utils     as ru
import radical.pilot     as rp
import radical.analytics as ra
import radical.entk as re


# In[13]:


# Input constants
trials = 2
stages = [16, 64, 256]
resource = 'bw'
src = '../raw-data/'
proc = '../proc-data/'


# In[3]:


def get_adap_time(loc, sid):
    
    # Get adap time
    duration = 0.0
    sess = ra.Session(stype='radical.entk', src=loc, sid=sid)
    stages = sorted(sess.filter(etype='stage', inplace=False).list('uid'))
#     print stages
    for stage in stages:
        duration += sess.duration(event=[{ru.EVENT: 'executing post-exec for stage %s'%stage},
                                         {ru.EVENT: 'post-exec executed for stage %s'%stage}])
    return duration


# In[4]:


def get_entk_overheads(loc, sid):
    
    sess = ra.Session(stype='radical.entk', src=loc, sid=sid)
    init_time = sess.duration(event=[{ru.EVENT: 'create amgr obj'},
                                     {ru.EVENT: 'init rreq submission'}])
    res_sub_time = sess.duration(event=[{ru.EVENT: 'creating rreq'},
                                     {ru.EVENT: 'rreq submitted'}])
    total_teardown_time = sess.duration(event=[{ru.EVENT: 'start termination'},
                                               {ru.EVENT: 'termination done'}])
    rts_teardown_time = sess.duration(event=[{ru.EVENT: 'canceling resource allocation'},
                                             {ru.EVENT: 'resource allocation cancelled'}])
    
    return {'init_time': init_time,
           'res_sub_time': res_sub_time,
           'total_teardown_time': total_teardown_time,
           'rts_teardown_time': rts_teardown_time}


# In[5]:


def get_entk_exec_time(loc, sid):
    sess = ra.Session(stype='radical.entk', src=loc, sid=sid)
    tasks = sess.filter(etype='task', inplace=False)
    return tasks.duration(state=['SCHEDULING','DONE'])


# In[6]:


def process_entk_profiles(src):
    
    sid = os.path.basename(src)
    loc = os.path.dirname(src)
    tag = '/'.join(loc.split('/')[2:])
    proc_data = os.path.join(proc,tag) + '/entk_data.json'
    data = {'adap_time': 0, 'overheads': 0, 'exec_time': 0}
        
    data['adap_time'] = get_adap_time(loc, sid)
    data['overheads'] = get_entk_overheads(loc, sid)
    data['exec_time'] = get_entk_exec_time(loc, sid)
    
    write_data(data, proc_data)
    return proc_data


# In[52]:


def write_data(data, proc_path):

    if 'rp.session' in proc_path:
        proc_path = os.path.dirname(os.path.dirname(proc_path)) + '/' + os.path.basename(proc_path)
    if not os.path.isdir(os.path.dirname(proc_path)):
        os.makedirs(os.path.dirname(proc_path))
    ru.write_json(data,proc_path)
    
    return proc_path


# In[51]:


print('EnTK analysis')
for s in stages:
    for t in range(1,trials+1):
        path = os.path.join(src,resource,'trial-%s'%t,'stages-%s'%s)
        for sess in glob(path + '/' + 're.session.*'):  
            print('Processing: ', sess)
            out_path = process_entk_profiles(sess)
            print('Output written to ', out_path)


# In[55]:


def process_rp_profiles(src):
    
    sid = os.path.basename(src)
    loc = os.path.dirname(src)
    tag = '/'.join(loc.split('/')[2:])
    proc_data = os.path.join(proc,tag) + '/rp_data.json'
    data = {'task_mgmt': 0, 'exec_time': 0}
    
    sess = ra.Session(stype='radical.pilot', src=loc, sid=sid)
    units = sess.filter(etype='unit', inplace=False)
        
    data['task_mgmt'] = units.duration(state=['NEW','DONE'])
    data['exec_time'] = units.duration(event=[{ru.EVENT:'exec_start'},{ru.EVENT:'exec_stop'}])
    
    proc_path = write_data(data, proc_data)
    return proc_path


# In[56]:


print('RP analysis')
for s in stages:
    for t in range(1,trials+1):
        path = os.path.join(src,resource,'trial-%s'%t,'stages-%s'%s)
        for sess in glob(path + '/' + 'rp.session.*/'):  
            print('Processing: ', sess)
            out_path = process_rp_profiles(sess)
            print('Output written to ', out_path)




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import os
import sys
import glob
import pprint
import radical.utils as ru
import radical.pilot as rp
import radical.analytics as ra

"""
This example illustrates how to obtain durations for arbitrary (non-state)
profile events.
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) != 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src     = sys.argv[1]
    session = ra.Session.create(src, 'radical')

    units = session.filter(etype='unit', inplace=False)
    print('#units   : %d' % len(units.get()))

    ranges = units.ranges(state=[['NEW'], ['DONE']], collapse=False)
    print('ranges   :')
    for r in ranges[:10]:
        print('  [%7.2f, %7.2f] = %7.2f' % (r[0], r[1], r[1] - r[0]))

    duration = units.duration(ranges=ranges)
    print('duration : %.2f' % duration)

    print("concurrent units in between exec_start and exec_stop events")
    concurrency = units.concurrency(state=[['NEW'], ['DONE']], sampling=10)
    pprint.pprint(concurrency[:10])


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys
import pprint

import radical.utils     as ru
import radical.analytics as ra

import matplotlib.pyplot as plt
import numpy             as np


# We plot all events listed below, but move the baseline to the first event.
# This will effectively plot the time spent in between the first and last event,
# with all intermediate events showing individual contributions (similar to
# a stacked bar plot).
#
# In the specific case examplified here, we plot the total duration of core
# allocation for each individual CU, including what events happen during that
# utilized time.  A core is considered allocated when a unit got sucessfully
# scheduled on it (then the core is blocked from use by other units), until the
# scheduler learns about the unit being finihed (then the core can be allocated
# to other units).
#
event_entity = 'unit'
event_list   = \
    [
        # {ru.STATE: 'NEW'                          , ru.EVENT: 'state'               },
        # {ru.STATE: 'UMGR_SCHEDULING_PENDING'      , ru.EVENT: 'state'               },
        # {ru.STATE: 'UMGR_SCHEDULING'              , ru.EVENT: 'state'               },
        # {ru.STATE: 'UMGR_STAGING_INPUT_PENDING'   , ru.EVENT: 'state'               },
        # {ru.STATE: 'UMGR_STAGING_INPUT'           , ru.EVENT: 'state'               },
        # {ru.STATE: 'AGENT_STAGING_INPUT_PENDING'  , ru.EVENT: 'state'               },
      #   {ru.COMP : 'agent_0'                      , ru.EVENT: 'get'                 },
      #   {ru.STATE: 'AGENT_STAGING_INPUT'          , ru.EVENT: 'state'               },
      #   {ru.STATE: 'AGENT_SCHEDULING_PENDING'     , ru.EVENT: 'state'               },
          {ru.STATE: 'AGENT_SCHEDULING'             , ru.EVENT: 'state'               },
          {ru.STATE: None                           , ru.EVENT: 'schedule_ok'         },
        # {ru.STATE: 'AGENT_EXECUTING_PENDING'      , ru.EVENT: 'state'               },
      #   {ru.STATE: 'AGENT_EXECUTING'              , ru.EVENT: 'state'               },
      #   {ru.STATE: None                           , ru.EVENT: 'exec_mkdir'          },
      #   {ru.STATE: None                           , ru.EVENT: 'exec_mkdir_done'     },
      #   {ru.STATE: None                           , ru.EVENT: 'exec_start'          },
        # {ru.STATE: None                           , ru.EVENT: 'exec_ok'             },
    #     {ru.STATE: None                           , ru.EVENT: 'cu_start'            },
      #   {ru.STATE: None                           , ru.EVENT: 'cu_cd_done'          },
    #     {ru.STATE: None                           , ru.EVENT: 'cu_exec_start'       },
      #   {ru.STATE: None                           , ru.EVENT: 'app_start'           },
    #     {ru.STATE: None                           , ru.EVENT: 'app_stop'            },
    #     {ru.STATE: None                           , ru.EVENT: 'cu_exec_stop'        },
      #   {ru.STATE: None                           , ru.EVENT: 'exec_stop'           },
        # {ru.STATE: None                           , ru.EVENT: 'unschedule_start'    },
      #   {ru.STATE: None                           , ru.EVENT: 'unschedule_stop'     },
        # {ru.STATE: 'AGENT_STAGING_OUTPUT_PENDING' , ru.EVENT: 'state'               },
      #   {ru.STATE: 'UMGR_STAGING_OUTPUT_PENDING'  , ru.EVENT: 'state'               },
        # {ru.STATE: 'UMGR_STAGING_OUTPUT'          , ru.EVENT: 'state'               },
        # {ru.STATE: 'AGENT_STAGING_OUTPUT'         , ru.EVENT: 'state'               },
        # {ru.STATE: 'DONE'                         , ru.EVENT: 'state'               },

    ]

log = False  # use log scale for y axis

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) != 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src     = sys.argv[1]
    session = ra.Session.create(src, 'radical.pilot')

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    session.filter(etype=event_entity, inplace=True)
    print('#entities: %d' % len(session.get()))

    data = dict()
    for thing in session.get():

        tstamps = list()
        for event in event_list:
            times = thing.timestamps(event=event)
            for e in thing.events:
                print(e)
            if times: tstamps.append(times[0])
            else    : tstamps.append(None)

        data[thing.uid] = tstamps

    # We sort the entities by the timestamp of the first event
    # Also, for all timestamps, we move the baseline to the first
    # timestamp in the list
    sorted_things = sorted(list(data.items()), key=lambda e: e[1][0])
    sorted_data   = list()
    index         = 0
  # for uid,tstamps in sorted_things[15:25]:
    for uid,tstamps in sorted_things:

        # rebase
        t_zero = tstamps[0]
        print(tstamps)
        for i in range(len(tstamps)):
            if tstamps[i]:
                tstamps[i] = tstamps[i] - t_zero
            else:
                print('pass', uid)

        # create plottable data
        sorted_data.append([index] + tstamps)
        index += 1


    # create a numpyarray for plotting
    np_data = np.array(sorted_data)
  # print np_data

    plt.figure(figsize=(20,14))
    for e_idx in range(len(event_list)):
        plt.scatter(np_data[:,0], np_data[:,(1+e_idx)], s=3, label=event_list[e_idx])

    if log:
        plt.yscale('log')
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15),
          ncol=2, fancybox=True, shadow=True)
    plt.savefig('08_core_utilization.png')
    plt.show()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

import sys
import pprint

import radical.utils as ru
import radical.pilot as rp
import radical.analytics as ra

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'

"""
This example illustrates the use of the method ra.Session.get()
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]

    if len(sys.argv) == 2: stype = 'radical.pilot'
    else                 : stype = sys.argv[2]

    session = ra.Session.create(src, stype)


    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    # and here we go. As seen in example 01, we use ra.Session.list() to get the
    # name of all the types of entity of the session.
    etypes = session.list('etype')
    pprint.pprint(etypes)

    # We limit ourselves to the types 'unit' and 'pilot'. We use the method
    # ra.Session.get() to get all the objects in our session with etype 'unit':
    ppheader("properties of the entities with etype 'unit'")
    units = session.get(etype='unit')
    pprint.pprint(units)

    # and 'pilot':
    ppheader("properties of the entities with etype 'pilot'")
    pilots = session.get(etype='pilot')
    pprint.pprint(pilots)

    # Mmmm, still a bit too many entities. We limit our analysis to a single
    # unit and pilot. We use ra.Session.get() to select all the objects in the
    # session with etype 'units' and uid 'unit.000000' and return them into a
    # list:
    ppheader("properties of the entities with etype 'unit' and uid 'unit.000000'")
    unit = session.get(etype='unit', uid='unit.000000')
    pprint.pprint(unit)

    # Because the uid is guaranteed to be unique within the scope of our
    # session, we can omit to specify etype, obtaining the same list as a
    # result:
    ppheader("properties of the entities with uid 'unit.000000'")
    unit = session.get(uid='unit.000000')
    pprint.pprint(unit)

    # We may want also to look into the states of this unit:
    ppheader("states of the entities with uid 'unit.000000'")
    states = unit[0].states
    pprint.pprint(states)

    # and extract the state we need. For example, the state 'NEW', that
    # indicates that the unit has been created. To refer to the state 'NEW', and
    # to all the other states of RADICAL-Pilot, we use the rp.NEW property that
    # guarantees type checking.
    ppheader("Properties of the state rp.NEW of the entities with uid 'unit.000000'")
    state = unit[0].states[rp.NEW]
    pprint.pprint(state)

    # Finally, we extract a property we need from this state. For example, the
    # timestamp of when the unit has been created, i.e., the property 'time' of
    # the state NEW:
    ppheader("Property 'time' of the state rp.NEW of the entities with uid 'unit.000000'")
    timestamp = unit[0].states[rp.NEW][ru.TIME]
    pprint.pprint(timestamp)

    # ra.Session.get() can also been used to to get all the entities in our
    # session that have a specific state. For example, the following gets all
    # the types of entity that have the state 'NEW':
    ppheader("Entities with state rp.NEW")
    entities = session.get(state=rp.NEW)
    pprint.pprint(entities)

    # We can then print the timestamp of the state 'NEW' for all the entities
    # having that state by using something like:
    ppheader("Timestamp of all the entities with state rp.NEW")
    timestamps = [entity.states[rp.NEW][ru.TIME] for entity in entities]
    pprint.pprint(timestamps)

    # We can also create tailored data structures for our analyis. For
    # example, using tuples to name entities, state, and timestamp:
    ppheader("Named entities with state rp.NEW and its timestamp")
    named_timestamps = [(entity.uid,
                         entity.states[rp.NEW][ru.STATE],
                         entity.states[rp.NEW][ru.TIME]) for entity in entities]
    pprint.pprint(named_timestamps)

    sys.exit(0)



#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys
import pprint

import radical.utils     as ru
import radical.analytics as ra

import matplotlib.pyplot as plt
import numpy             as np


# We look into individual contributions of sub-durations to a larger duration.
# The event list below will describe the whole duration (first to last event),
# and the durations between subsequential events are considered contributing
# sub-durations.  For each entity, we plot the times derived that way.
#
event_entity = 'unit'
event_list   = \
    [
    # {ru.STATE: 'NEW'                          , ru.EVENT: 'state'           },
    # {ru.STATE: 'UMGR_SCHEDULING_PENDING'      , ru.EVENT: 'state'           },
    # {ru.STATE: 'UMGR_SCHEDULING'              , ru.EVENT: 'state'           },
    # {ru.STATE: 'UMGR_STAGING_INPUT_PENDING'   , ru.EVENT: 'state'           },
    # {ru.STATE: 'UMGR_STAGING_INPUT'           , ru.EVENT: 'state'           },
    # {ru.STATE: 'AGENT_STAGING_INPUT_PENDING'  , ru.EVENT: 'state'           },
    # {ru.STATE: None                           , ru.EVENT: 'get'             },
    # {ru.STATE: 'AGENT_STAGING_INPUT'          , ru.EVENT: 'state'           },
    # {ru.STATE: 'AGENT_SCHEDULING_PENDING'     , ru.EVENT: 'state'           },
      {ru.STATE: 'AGENT_SCHEDULING'             , ru.EVENT: 'state'           },
    # {ru.STATE: None                           , ru.EVENT: 'schedule_ok'     },
      {ru.STATE: 'AGENT_EXECUTING_PENDING'      , ru.EVENT: 'state'           },
      {ru.STATE: 'AGENT_EXECUTING'              , ru.EVENT: 'state'           },
      {ru.STATE: None                           , ru.EVENT: 'exec_start'      },
    # {ru.STATE: None                           , ru.EVENT: 'exec_ok'         },
      {ru.STATE: None                           , ru.EVENT: 'exec_stop'       },
    # {ru.STATE: None                           , ru.EVENT: 'unschedule_start'},
      {ru.STATE: None                           , ru.EVENT: 'unschedule_stop' },
    # {ru.STATE: 'AGENT_STAGING_OUTPUT_PENDING' , ru.EVENT: 'state'           },
    # {ru.STATE: 'UMGR_STAGING_OUTPUT_PENDING'  , ru.EVENT: 'state'           },
    # {ru.STATE: 'UMGR_STAGING_OUTPUT'          , ru.EVENT: 'state'           },
    # {ru.STATE: 'AGENT_STAGING_OUTPUT'         , ru.EVENT: 'state'           },
    # {ru.STATE: 'DONE'                         , ru.EVENT: 'state'           },
    ]

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) != 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src     = sys.argv[1]
    session = ra.Session.create(src, 'radical.pilot')

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print(separator + message + separator)

    session.filter(etype=event_entity, inplace=True)
    print('#entities: %d' % len(session.get()))

    data = dict()
    for thing in session.get():

        tstamps = list()

        for event in event_list:
            times = thing.timestamps(event=event)
            if times: tstamps.append(times[0])
            else    : tstamps.append(np.nan)

        data[thing.uid] = tstamps

    # We sort the entities by the timestamp of the first event
    # We also derive the durations, first the individual contributions, then the
    # overall duration.
    # timestamp in the list
    sorted_things = sorted(list(data.items()), key=lambda e: e[1][0])
    sorted_data   = list()
    index         = 0
    for uid,tstamps in sorted_things:

        durations = list()
        durations.append(tstamps[-1] - tstamps[0])  # global duration
        for i in range(len(tstamps) - 1):
            durations.append(tstamps[i + 1] - tstamps[i])

        # create plottable data
        sorted_data.append([index] + durations)
        index += 1

    # create a numpyarray for plotting
    np_data = np.array(sorted_data)
  # print np_data

    plt.figure(figsize=(20,14))
    for e_idx in range(len(event_list)):
        if e_idx == 0:
            label = 'total'
        else:
            label = '%s - %s' % (ru.event_to_label(event_list[e_idx - 1]),
                                 ru.event_to_label(event_list[e_idx]))
        plt.plot(np_data[:,0], np_data[:,(1 + e_idx)], label=label)

    plt.yscale('log')
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15),
          ncol=2, fancybox=True, shadow=True)
    plt.savefig('08b_core_utilization.png')
    plt.show()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys

import matplotlib        as mpl
import matplotlib.pyplot as plt

import radical.utils     as ru
import radical.pilot     as rp
import radical.analytics as ra

states  = [
           [rp.NEW                          , '#660000'],
         # [rp.UMGR_SCHEDULING_PENDING      , '#000000'],
           [rp.UMGR_SCHEDULING              , '#666600'],
         # [rp.UMGR_STAGING_INPUT_PENDING   , '#000000'],
           [rp.UMGR_STAGING_INPUT           , '#006600'],
         # [rp.AGENT_STAGING_INPUT_PENDING  , '#000000'],
           [rp.AGENT_STAGING_INPUT          , '#006666'],
         # [rp.AGENT_SCHEDULING_PENDING     , '#000000'],
           [rp.AGENT_SCHEDULING             , '#000066'],
           [rp.AGENT_EXECUTING_PENDING      , '#000000'],
           [rp.AGENT_EXECUTING              , '#660066'],
         # [rp.AGENT_STAGING_OUTPUT_PENDING , '#000000'],
           [rp.AGENT_STAGING_OUTPUT         , '#990000'],
         # [rp.UMGR_STAGING_OUTPUT_PENDING  , '#000000'],
           [rp.UMGR_STAGING_OUTPUT          , '#009900'],
           [rp.DONE                         , '#00CC00'],
           [rp.FAILED                       , '#FF0000'],
           [rp.CANCELED                     , '#AA00AA'],
          ]
# state to sort units by
state_key = rp.AGENT_EXECUTING
state_key = rp.NEW


concurrency = {
        'Unit Scheduling' : [{ru.STATE: 'AGENT_SCHEDULING'},
                             {ru.EVENT: 'schedule_ok'     }],
        'Unit Execution'  : [{ru.EVENT: 'exec_start'      },
                             {ru.EVENT: 'exec_stop'       }],
}

rate = {
       'Unit Scheduling' : {ru.EVENT: 'schedule_ok' },
       'Unit Execution'  : {ru.EVENT: 'exec_start'  },
}

colors = {'Unit Scheduling': '#CC5555',
          'Unit Execution':  '#55CC55'}


# ------------------------------------------------------------------------------
#
def plot_states(session):

    data = list()
    for task in session.get(etype='unit'):

        info = [task.uid]
        for state,_ in states:
            info.append(task.states.get(state, [None])[ru.TIME])
        data.append(info)

    sort_idx = 0
    for state,_ in states:
        sort_idx += 1
        if state == state_key:
            break

    data.sort(key=lambda x: x[sort_idx])

    # prep figure
    fig  = plt.figure(figsize=(20,14))
    ax   = fig.add_subplot(111)

    for idx,state in enumerate(states):
      # print len(list(range(len(data))))
      # print len([d[idx + 1] for d in data])

        plt.scatter([d[idx + 1] for d in data],
                    list(range(len(data))),
                    s=5,
                    color=state[1],
                    label=state[0])

    ax.legend(ncol=5, loc='upper center',
              bbox_to_anchor=(0.5,1.11))
    plt.xlabel('time [s]')
    plt.ylabel('#tasks')

    fig.savefig('%s_states.png' % session.uid)
  # plt.show()


# ------------------------------------------------------------------------------
#
def plot_rates(session):

    data = dict()
    for metric in rate:
        data[metric] = session.rate(event=rate[metric], sampling=0.1)

    # prep figure
    fig  = plt.figure(figsize=(20,14))
    ax   = fig.add_subplot(111)

    for metric in data:
        x = [e[0] for e in data[metric]]
        y = [e[1] for e in data[metric]]
        plt.step(x, y, color=colors[metric], label=metric, where='post',
                linewidth=2, alpha=0.8)

    ax.legend(list(data.keys()), ncol=3, loc='upper center',
                                 bbox_to_anchor=(0.5,1.11))
    plt.xlabel('time [s]')
    plt.ylabel('rate (#tasks / sec)')

    fig.savefig('%s_rates.png' % session.uid)
  # plt.show()


# ------------------------------------------------------------------------------
#
def plot_concurrency(session):

    data = dict()
    for metric in concurrency:
        data[metric] = session.concurrency(event=concurrency[metric])

    # prep figure
    fig  = plt.figure(figsize=(20,14))
    ax   = fig.add_subplot(111)

    for metric in data:
        x = [e[0] for e in data[metric]]
        y = [e[1] for e in data[metric]]
        plt.step(x, y, color=colors[metric], label=metric, where='post')

    ax.legend(list(data.keys()), ncol=3, loc='upper center',
                                 bbox_to_anchor=(0.5,1.11))
    plt.xlabel('time [s]')
    plt.ylabel('concurrency (#tasks)')

    fig.savefig('%s_concurrency.png' % session.uid)
  # plt.show()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) != 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src = sys.argv[1]
    session = ra.Session.create(src, 'radical.pilot')

    plot_states(session)
    plot_rates(session)
    plot_concurrency(session)


# ------------------------------------------------------------------------------






# ------------------------------------------------------------------------------
#
class Plotter(object):

    def __init__(self,style=None,plot_grid=None):
        '''
        Create a radical.analytics plot object.

        The plot is created to facilitate several types of plots for at least
        the well know data organizations that exist in radical.analytics, like
        timeseries, events, statistical plot (e.g. histograms, kernels, etc).

        The object is created by passing the following information:
        style         : The style of plot, i.e. matplotlib or bokeh #TODO
        plot_grid     : The plotting format. This is a list that keeps the
                        number of rows and columns of plots [nrows,ncols]
        '''

        self._style = style
        self._plot_grid = plot_grid

        try:
            import matplotlib.pyplot as plt
            self._plt = plt

        except:
            raise RuntimeError('Plotter class needs matplotlib installed')


    # --------------------------------------------------------------------------
    #
    def utilization(self, util_data=None, normalized=None,
                          resource=None, fig_size=None):
        '''
        This method receives as input utilization data of a specific resource
        and creates a figure for that specific utilization.

        The method is able to normalized the plot based on the owner resources
        and the utilized resources at any given time. In addition, figure's size
        can be provided as input.
        '''

        # Initially check if plot_grid is set. If not, create a single plot,
        # else create a figure based on the user's specification.
        if not self._plot_grid:

            self._fig,self._axis = self._plt.subplots(nrows=1, ncols=1,
                                                      figsize=fig_size)

            # Iterate over the owners and add a line for every owner.
            for key, util in list(util_data.items()):

                # Getting the time moments where the utilization changes
                x_axis = [point[0] for point in util['utilization']]

                # Getting the utilization points. If normalized flag is set,
                # utilization is divided with the total resources of the owner
                # and multiply with 100
                if normalized:
                    y_axis = [(point[1] / util['resources']) * 100
                                       for point in util['utilization']]
                else:
                    y_axis = [point[1] for point in util['utilization']]

                # If there is a range where the owner had the resources get it.
                # Otherwise it is none. The resource range will be use to set
                # the utilization plot correctly in the x-axis
                resource_range = util['range']
                if not resource_range:
                    resource_range = None

                if resource_range:
                    self._axis.set_xlim(resource_range)
                    self._axis.plot([resource_range[0] + x for x in x_axis],
                                    y_axis, label='%s utilization' % key)
                else:
                    self._axis.plot(x_axis,y_axis,label='%s utilization' % key)

                # X,Y label, and title are being set
                self._axis.set_ylabel('Utilization in Resources')
                self._axis.set_xlabel('Time in seconds')
                self._axis.set_title('%s Resource Utilization' % resource)
                self._axis.legend()

        else:
            self._fig,self._axis = self._plt.subplots(nrows=self._plot_grid[0],
                                                      ncols=self._plot_grid[1],
                                                      figsize=fig_size)

            # FIXME: this code is too obscure
            for (key, util),i \
                in zip(iter(list(util_data.items())), 
                   list(range(self._plot_grid[0] * self._plot_grid[1]))):

                # Getting the time moments where the utilization changes
                x_axis = [point[0] for point in util['utilization']]

                # Getting the utilization points. If normalized flag is set, utilization is
                # divided with the total resources of the owner and multiply with 100
                if normalized:
                    y_axis = [(point[1] / util['resources']) * 100 
                                       for point in util['utilization']]
                else:
                    y_axis = [point[1] for point in util['utilization']]

                # If there is a range where the owner had the resources get it. Otherwise
                # it is none. The resource range will be use to set the utilization plot
                # correctly in the x-axis
                resource_range = util['range'] if util['range'] else 0

                if resource_range:
                    self._axis[i].set_xlim(resource_range)
                    self._axis[i].plot([resource_range[0] + x for x in x_axis],
                                       y_axis, label='%s utilization' % key)
                else:
                    self._axis[i].plot(x_axis,y_axis,
                                       label='%s utilization' % key)

                # X,Y label, and title are being set
                self._axis[i].set_ylabel('Utilization in Resources')
                self._axis[i].set_xlabel('Time in seconds')
                self._axis[i].set_title('%s Resource Utilization' % resource)
                self._axis[i].legend()


    # --------------------------------------------------------------------------
    #
    def concurrency(self,data=None,fig_size=None):
        '''
        This method receives as input concurrency data and creates a figure.
        The user is able to set the figure size.
        '''

        self._fig,self._axis = self._plt.subplots(nrows=1, ncols=1,
                                                  figsize=fig_size)

        # The x axis is setup.
        x_axis = [point[0] for point in data]

        # The y axis is setup.
        y_axis = [point[1] for point in data]

        # Just plot
        self._axis.plot(x_axis,y_axis)        
        self._axis.set_ylabel('Concurrent Entities')
        self._axis.set_xlabel('Time in seconds')
        self._axis.set_title('Concurrency')


# ------------------------------------------------------------------------------





import os
import json
import pytest
import radical.utils as ru
from radical.analytics.entity import Entity


# Test Directory use to load example json files
base = "%s/example-data" % os.path.dirname(__file__)


# ------------------------------------------------------------------------------
#
@pytest.fixture
def pilot_entity():
    """Fixture to get the example Pilot entity from example data"""

    with ru.ru_open("%s/pilot-entity-example.json" % base, 'r') as f:
        entity = json.load(f)
        entity['events'] = [tuple(event) for event in entity['events']]
        return entity


# ------------------------------------------------------------------------------
#
@pytest.fixture
def range_entity():
    """Fixture to get the example range-testing entity from example data"""

    with ru.ru_open("%s/range-testing-entity-example.json" % base, 'r') as f:
        entity = json.load(f)
        entity['events'] = [tuple(event) for event in entity['events']]
        return entity


# ------------------------------------------------------------------------------
#
def get_states(events=None):
    """Get a dictionary of states from a list of event tuples"""

    ret_states = dict()
    if events is not None:
        for event in events:
            if event[ru.EVENT] == 'state':
                ret_states[event[ru.STATE]] = event

    return ret_states


# ------------------------------------------------------------------------------
#
def sort_events(events=None):
    """Sort a list of tuples of events"""

    ret_events = list()
    if events is not None:
        for event in sorted(events, key=lambda x: (x[ru.TIME])):
            ret_events.append(event)

    return ret_events


# ------------------------------------------------------------------------------
#
class TestEntity(object):

    # --------------------------------------------------------------------------
    #
    def test_t_start(self, pilot_entity):
        """Test a valid t_start"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )
        events = sort_events(pilot_entity['events'])

        # Should match the time in the first element after sorting
        assert e.t_start == events[0][ru.TIME]


    # --------------------------------------------------------------------------
    #
    def test_t_stop(self, pilot_entity):
        """Test a valid t_stop"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )
        events = sort_events(pilot_entity['events'])

        # Should match the time in the last element after sorting
        assert events[-1][ru.TIME] == e.t_stop


    # --------------------------------------------------------------------------
    #
    def test_ttc(self, pilot_entity):
        """Test a valid ttc"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )
        events = sort_events(pilot_entity['events'])

        # Should match the difference in time between
        # first and last elements after sorting
        assert e.ttc == float(events[-1][ru.TIME] - events[0][ru.TIME])


    # --------------------------------------------------------------------------
    #
    def test_t_range(self, pilot_entity):
        """Test a valid t_range"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )
        events = sort_events(pilot_entity['events'])

        # Should match exactly what was passed in...
        assert [events[0][ru.TIME], events[-1][ru.TIME]] == e.t_range


    # --------------------------------------------------------------------------
    #
    def test_uid(self, pilot_entity):
        """Test a valid uid"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )

        # Should match exactly what was passed in...
        assert e.uid == pilot_entity['uid']


    # --------------------------------------------------------------------------
    #
    def test_etype(self, pilot_entity):
        """Test a valid etype"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )

        # Should match exactly what was passed in...
        assert e.etype == pilot_entity['etype']


    # --------------------------------------------------------------------------
    #
    def test_states(self, pilot_entity):
        """Test a valid states"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )
        # Get all events in the example file and put them in a dictionary
        states = get_states(pilot_entity['events'])

        # Should match a list of events of type 'state'
        # extracted from the events passed in
        assert e.states == states


    # --------------------------------------------------------------------------
    #
    def test_events(self, pilot_entity):
        """Test a valid events"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )
        # Get all events in order by timestamp
        events = sort_events(pilot_entity['events'])

        # Should match a list of sorted events (by the time field)
        # generated from the list of events passed in
        assert e.events == events


    # --------------------------------------------------------------------------
    #
    def test_description(self, pilot_entity):
        """Test a valid description"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )

        # Should match exactly what was passed in...
        assert type(e.description) is dict
        assert e.description == pilot_entity['details']['description']


    # --------------------------------------------------------------------------
    #
    def test_cfg(self, pilot_entity):
        """Test a valid cfg"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )

        # Should match exactly what was passed in...
        assert type(e.cfg) is dict
        assert e.cfg == pilot_entity['details']['cfg']


    # --------------------------------------------------------------------------
    #
    def test_consistency(self, pilot_entity):
        """Test a valid consistency"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )

        # TODO: Better matching, for now we match just a dict()
        assert type(e.consistency) is dict


    # --------------------------------------------------------------------------
    #
    def test_as_dict(self, pilot_entity):
        """Test a valid as_dict"""
        e = Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )
        expected = {
            'uid': pilot_entity['uid'],
            'etype': pilot_entity['etype'],
            'states': get_states(pilot_entity['events']),
            'events': sort_events(pilot_entity['events'])
        }

        # Should match exactly what was passed in
        # as the order seen above...
        # don't match `cfg` and `description` as those are runtime dependent
        edict = e.as_dict()
        del edict['cfg']
        del edict['description']
        del edict['resources']
        assert edict == expected


    ##############################################################
    # Test Ranges Method: expand=False, collapse=True
    #
    # We match range values form the example data at:
    #   ./example-data/range-testing-entity-example.json
    ##############################################################

    # --------------------------------------------------------------------------
    #
    def test_ranges_one_state(self, range_entity):
        """Test a valid ranges result with one state start/finish"""
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Only one range of one start/end
        ranges = e.ranges(state=['PMGR_ACTIVE_PENDING', 'FAILED'])
        assert ranges == [[21.668299913406372, 50.227399826049805]]


    # --------------------------------------------------------------------------
    #
    def test_ranges_two_consecutive_state(self, range_entity):
        """Test a valid ranges result with two states start/finish"""
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Two consecutive ranges which get collapsed to the spanning max
        ranges = e.ranges(state=[['NEW',            'PMGR_ACTIVE'     ],
                                 ['PMGR_LAUNCHING', 'PMGR_ACTIVE_PENDING']])
        assert ranges == [[0.0, 4.449499845504761]]


    # --------------------------------------------------------------------------
    #
    def test_ranges_time_filter_exact_match(self, range_entity):
        """Test a valid ranges with time filter, exact matches"""
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # for state
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [21.668299913406372, 50.227399826049805]
        ])
        assert ranges == [
            [21.668299913406372, 50.227399826049805]
        ]

        # for events
        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [4.446699857711792, 29.150099992752075]
        ])
        assert ranges == [
            [4.446699857711792, 29.150099992752075]
        ]


    # --------------------------------------------------------------------------
    #
    def test_ranges_time_filter_no_match(self, range_entity):
        """Test a valid ranges with time filter, exact matches"""
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # for state
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [60.0, 70.0]
        ])
        assert ranges == []

        # for events
        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [30.0, 40.0]
        ])
        assert ranges == []


    # --------------------------------------------------------------------------
    #
    def test_ranges_time_filter_intersection_match(self, range_entity):
        """Test a valid ranges with time filter, one intersection match each"""
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # As you read the tests, here is what each symbol represents:
        #   [      ] --> matched range before time filtering
        #   {      } --> time filtering value
        #   (      ) --> matched range *after* time filtering

        # Inner intersection: [   {(  )}   ]
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [25.0, 30.0]
        ])
        assert ranges == [
            [25.0, 30.0]
        ]

        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [5.0, 25.0]
        ])
        assert ranges == [
            [5.0, 25.0]
        ]

        # Outter intersection: {   ([    ])   }
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [10.0, 60.0]
        ])
        assert ranges == [
            [21.668299913406372, 50.227399826049805]
        ]

        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [0.0, 50.0]
        ])
        assert ranges == [
            [4.446699857711792, 29.150099992752075]
        ]

        # Right-side intersection match: {   ([   )}   ]
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [25.0, 100.00]
        ])
        assert ranges == [
            [25.0, 50.227399826049805]
        ]

        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [5.0, 50.0]
        ])
        assert ranges == [
            [5.0, 29.150099992752075]
        ]

        # Left-side intersection match: [   {(   ])   }
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [0, 30.0]
        ])
        assert ranges == [
            [21.668299913406372, 30.0]
        ]

        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [0.0, 25.0]
        ])
        assert ranges == [
            [4.446699857711792, 25.0]
        ]


    # --------------------------------------------------------------------------
    #
    def test_ranges_multi_time_filter_exact_match(self, range_entity):
        """Test a valid ranges with multiple time filters,
        exact matches only"""
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # As you read the tests, here is what each symbol represents:
        #   [      ] --> matched range before time filtering
        #   {      } --> time filter #1
        #   |      | --> time filter #2
        #   (      ) --> matched range #1 *after* time filtering
        #   <      > --> matched range #2 *after* time filtering

        # Only one of the filters match: {([   ])}   |   |
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [21.668299913406372, 50.227399826049805],
            [60.0, 80.0]
        ])
        assert ranges == [
            [21.668299913406372, 50.227399826049805]
        ]

        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [4.446699857711792, 29.150099992752075],
            [30.0, 40.0]
        ])
        assert ranges == [
            [4.446699857711792, 29.150099992752075]
        ]

        # Only one of the filters match: {   }   |<[   ]>|
        ranges = e.ranges(state=[
            ['PMGR_ACTIVE_PENDING'],
            ['FAILED']
        ], time=[
            [60.0, 80.0],
            [21.668299913406372, 50.227399826049805]
        ])
        assert ranges == [
            [21.668299913406372, 50.227399826049805]
        ]

        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'}],
            [{ru.EVENT: 'sync_rel'}]
        ], time=[
            [30.0, 40.0],
            [4.446699857711792, 29.150099992752075]
        ])
        assert ranges == [
            [4.446699857711792, 29.150099992752075]
        ]


    # --------------------------------------------------------------------------
    #
    def test_ranges_multi_time_filter_intersection_match(
            self, range_entity):
        """Test a valid ranges with multiple time filters,
            intersection matches each"""
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        assert e.t_start       ==  0
        assert e.t_stop        ==     50.507999897003174
        assert e.ttc           ==     50.507999897003174
        assert e.t_range       == [0, 50.507999897003174]
        assert len(e.list_states()) ==  6
        assert len(e.states       ) ==  6
        assert len(e.events       ) == 22
        # TODO: Add more assertions as needed

        # As you read the tests, here is what each symbol represents:
        #   [      ] --> matched range before time filtering
        #   {      } --> time filter #1
        #   |      | --> time filter #2
        #   (      ) --> matched range #1 *after* time filtering
        #   <      > --> matched range #2 *after* time filtering

        # Non-overlaping filters, one matches:    {([   )}   ]   |   |
        ranges = e.ranges(state=['PMGR_ACTIVE_PENDING', 'FAILED'],
                          time=[[21.668299913406372, 25.0], [60.0, 70.0]])
        assert ranges == [[21.668299913406372, 25.0]]

        ranges = e.ranges(event=[{ru.EVENT: 'put'}, {ru.EVENT: 'sync_rel'}],
                          time=[[4.446699857711792, 15.0], [30.0, 50.0]
        ])
        assert ranges == [[4.446699857711792, 15.0]]

        # Non-overlaping filters, both match: {   ([   )}   |<   ]>   |
        ranges = e.ranges(state=['PMGR_ACTIVE_PENDING', 'FAILED'],
                          time=[[10, 25.0], [30.0, 70.0]])
        assert ranges == [[21.668299913406372, 25.0],
                           [30.0, 50.227399826049805]]

        ranges = e.ranges(event=[{ru.EVENT: 'put'}, {ru.EVENT: 'sync_rel'}],
                          time=[[0.0, 15.0], [20.0, 30.0]])
        assert ranges == [[4.446699857711792, 15.0],
                           [20.0, 29.150099992752075]]

        # semi-overlaping filters, one matches: [   {(   ])   |   }   |
        ranges = e.ranges(state=['PMGR_ACTIVE_PENDING', 'FAILED'],
                          time=[[25.0, 70.0], [60.0, 80.0]])
        assert ranges == [[25.0, 50.227399826049805]]

        ranges = e.ranges(event=[{ru.EVENT: 'put'}, {ru.EVENT: 'sync_rel'}],
                          time=[[15.0, 40.0], [30.0, 50.0]])
        assert ranges == [[15.0, 29.150099992752075]]

        # semi-overlaping filters, both match: {   ([   |<   )}   ]>   |
        ranges = e.ranges(state=['PMGR_ACTIVE_PENDING', 'FAILED'],
                          time=[[10, 30.0], [25.0, 70.0]])
        assert ranges == [[21.668299913406372, 50.227399826049805]]

        ranges = e.ranges(event=[{ru.EVENT: 'put'}, {ru.EVENT: 'sync_rel'}],
                          time=[[0.0, 20.0], [15.0, 30.0]])
        assert ranges == [[4.446699857711792, 29.150099992752075]]

        # complete-overlap filters, both match: |   <[   {(   )}   ]>   |
        ranges = e.ranges(state=['PMGR_ACTIVE_PENDING', 'FAILED'],
                          time=[[25.0, 35.0], [10.0, 70.0]])
        assert ranges == [[21.668299913406372, 50.227399826049805]]

        ranges = e.ranges(event=[{ru.EVENT: 'put'}, {ru.EVENT: 'sync_rel'}],
                          time=[[10.0, 20.0], [0.0, 30.0]])
        assert ranges == [[4.446699857711792, 29.150099992752075]]


    # --------------------------------------------------------------------------
    #
    def test_ranges_two_overlaping_state(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Two overlaping ranges
        ranges = e.ranges(state=[['NEW', 'PMGR_ACTIVE_PENDING'],
                                 ['PMGR_LAUNCHING', 'PMGR_ACTIVE']])
        assert ranges == [[4.433599948883057, 4.449499845504761]]


    # --------------------------------------------------------------------------
    #
    def test_ranges_one_event(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Only one range of one start/end
        ranges = e.ranges(event=[
            ({ru.EVENT: 'put'}),
            ({ru.EVENT: 'cmd'})
        ])
        assert ranges == [
            [4.446699857711792, 50.507999897003174]
        ]


    # --------------------------------------------------------------------------
    #
    def test_ranges_one_event_multi_match(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Only one range of one start/end
        ranges = e.ranges(event=[
            ({ru.EVENT: 'update_request'}),
            ({ru.EVENT: 'update_pushed'})
        ])
        assert ranges == [
            [4.447200059890747 , 4.457900047302246],
            [4.458099842071533 , 5.236199855804443],
            [21.666899919509888, 21.685499906539917],
            [29.752099990844727, 29.939599990844727],
            [30.722899913787842, 32.28690004348755 ],
            [50.22889995574951 , 50.240999937057495]
        ]


    # --------------------------------------------------------------------------
    #
    def test_ranges_two_consecutive_event(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Two consecutive ranges
        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'     }, {ru.EVENT: 'hostname'}],
            [{ru.EVENT: 'sync_rel'}, {ru.EVENT: 'cmd'     }]
        ])
        print(ranges)
        assert ranges == [[4.446699857711792, 29.150099992752075],
                          [29.760799884796143, 50.507999897003174]]


    # --------------------------------------------------------------------------
    #
    def test_ranges_two_overlaping_event(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # FIXME: this is the same test as above

        # Two overlaping ranges
        ranges = e.ranges(event=[
            [{ru.EVENT: 'put'     }, {ru.EVENT: 'hostname'}],
            [{ru.EVENT: 'sync_rel'}, {ru.EVENT: 'cmd'     }]
        ])
        assert ranges == [
            [4.446699857711792, 29.150099992752075],
            [29.760799884796143, 50.507999897003174]
        ]


    # --------------------------------------------------------------------------
    #
    def test_empty_profile(self, pilot_entity):

        with pytest.raises(Exception):
            Entity(_uid=pilot_entity['uid'],
                   _profile=list(),
                   _details=pilot_entity['details'])


    ##############################################################
    # Test invalid scenarios
    ##############################################################

    # --------------------------------------------------------------------------
    #
    def test_none_uid(self, pilot_entity):

        with pytest.raises(Exception):
            Entity(_uid=None,
                   _profile=pilot_entity['events'],
                   _details=pilot_entity['details']
                   )


    # --------------------------------------------------------------------------
    #
    def test_none_etype(self, pilot_entity):

        Entity(_uid=pilot_entity['uid'],
               _profile=pilot_entity['events'],
               _details=pilot_entity['details']
               )


    # --------------------------------------------------------------------------
    #
    def test_none_profile(self, pilot_entity):
        with pytest.raises(Exception):
            Entity(_uid=None,
                   _profile=None,
                   _details=pilot_entity['details']
                   )


    # --------------------------------------------------------------------------
    #
    def test_none_details(self, pilot_entity):
        with pytest.raises(Exception):
            Entity(_uid=pilot_entity['uid'],
                   _profile=pilot_entity['events'],
                   _details=None
                   )


    # --------------------------------------------------------------------------
    #
    def test_invalid_ranges_no_end_matched_state(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Only one range of one start/end
        ranges = e.ranges(event=[({ru.EVENT: 'put'}), ({ru.EVENT: 'aaa'})
        ])
        assert ranges == []


    # --------------------------------------------------------------------------
    #
    def test_invalid_ranges_no_end_matched_event(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Only one range of one start/end
        ranges = e.ranges(event=[('NEW'), ('AAA') ])
        assert ranges == []


    # --------------------------------------------------------------------------
    #
    def test_invalid_ranges_no_start_matched_event(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Only one range of one start/end
        ranges = e.ranges(event=[({ru.EVENT: 'aaa'}), ({ru.EVENT: 'put'})
        ])
        assert ranges == []


    # --------------------------------------------------------------------------
    #
    def test_invalid_ranges_no_start_matched_state(self, range_entity):
        e = Entity(_uid=range_entity['uid'],
                   _profile=range_entity['events'],
                   _details=range_entity['details']
                   )

        # Only one range of one start/end
        ranges = e.ranges(event=[('AAA'), ('NEW')])
        assert ranges == []


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys

import matplotlib.pyplot as plt

import radical.utils     as ru
import radical.pilot     as rp
import radical.analytics as ra

from radical.analytics.utils import to_latex


# ----------------------------------------------------------------------------
#
plt.style.use(ra.get_mplstyle("radical_mpl"))


# ----------------------------------------------------------------------------
#
states  = [
           [rp.NEW                          , '#660000'],
         # [rp.TMGR_SCHEDULING_PENDING      , '#000000'],
           [rp.TMGR_SCHEDULING              , '#666600'],
         # [rp.TMGR_STAGING_INPUT_PENDING   , '#000000'],
           [rp.TMGR_STAGING_INPUT           , '#006600'],
         # [rp.AGENT_STAGING_INPUT_PENDING  , '#000000'],
           [rp.AGENT_STAGING_INPUT          , '#006666'],
         # [rp.AGENT_SCHEDULING_PENDING     , '#000000'],
           [rp.AGENT_SCHEDULING             , '#000066'],
           [rp.AGENT_EXECUTING_PENDING      , '#000000'],
           [rp.AGENT_EXECUTING              , '#660066'],
         # [rp.AGENT_STAGING_OUTPUT_PENDING , '#000000'],
           [rp.AGENT_STAGING_OUTPUT         , '#990000'],
         # [rp.TMGR_STAGING_OUTPUT_PENDING  , '#000000'],
           [rp.TMGR_STAGING_OUTPUT          , '#009900'],
           [rp.DONE                         , '#00CC00'],
           [rp.FAILED                       , '#FF0000'],
           [rp.CANCELED                     , '#AA00AA'],
          ]


metrics = {'Task Scheduling': {ru.EVENT: 'schedule_ok'},
           'Task Execution' : {ru.EVENT: 'exec_start' },
          }
colors  = {'Task Scheduling': '#CC5555',
           'Task Execution' : '#55CC55'}


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src     = sys.argv[1]
    stype   = 'radical.pilot'
    session = ra.Session.create(src, stype)

    # FIXME: adaptive sampling (100 bins over range?)
    data = {metric: session.rate(event=metrics[metric], sampling=1.0)
            for metric in metrics}

    fig, ax = plt.subplots(figsize=ra.get_plotsize(500))

    for metric in data:
        x = [e[0] for e in data[metric]]
        y = [e[1] for e in data[metric]]
        # FIXME: use cmap
        ax.plot(x, y, color=colors[metric], label=to_latex(metric))
      # ax.step(x, y, color=colors[metric], label=to_latex(metric),
      #               where='post', linewidth=2, alpha=0.8)

    ax.legend(to_latex(list(data.keys())))
    # FIXME: why is the x-axis label gone?
    plt.xlabel(to_latex('time [s]'))
    plt.ylabel(to_latex('rate (#tasks / sec)'))

    fig.savefig('%s_rate.png' % session.uid)


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import sys

import radical.utils     as ru
import radical.pilot     as rp
import radical.analytics as ra

import matplotlib.pyplot as plt
import numpy             as np

from radical.analytics.utils import to_latex


# ----------------------------------------------------------------------------
#
plt.style.use(ra.get_mplstyle("radical_mpl"))


# We plot timelines for all events listed in `event_list` for all entities of
# type `event_entity`..  Before plotting, we sort those entities by the
# timestamp of the first event in the event list

event_entities = ['task', 'master', 'worker']
event_list   = [
    # {ru.STATE: rp.NEW                         , ru.EVENT: 'state'          },
    # {ru.STATE: rp.TMGR_SCHEDULING_PENDING     , ru.EVENT: 'state'          },
    # {ru.STATE: rp.TMGR_SCHEDULING             , ru.EVENT: 'state'          },
    # {ru.STATE: rp.TMGR_STAGING_INPUT_PENDING  , ru.EVENT: 'state'          },
    # {ru.STATE: rp.TMGR_STAGING_INPUT          , ru.EVENT: 'state'          },
    # {ru.STATE: rp.AGENT_STAGING_INPUT_PENDING , ru.EVENT: 'state'          },
    # {ru.COMP : 'agent_0'                      , ru.EVENT: 'get'            },
      {ru.STATE: rp.AGENT_STAGING_INPUT         , ru.EVENT: 'state'          },
      {ru.STATE: rp.AGENT_SCHEDULING_PENDING    , ru.EVENT: 'state'          },
      {ru.STATE: rp.AGENT_SCHEDULING            , ru.EVENT: 'state'          },
    # {ru.STATE: None                           , ru.EVENT: 'schedule_ok'    },
      {ru.STATE: rp.AGENT_EXECUTING_PENDING     , ru.EVENT: 'state'          },
      {ru.STATE: rp.AGENT_EXECUTING             , ru.EVENT: 'state'          },
    # {ru.STATE: None                           , ru.EVENT: 'exec_mkdir'     },
    # {ru.STATE: None                           , ru.EVENT: 'exec_mkdir_done'},
      {ru.STATE: None                           , ru.EVENT: 'exec_start'     },
    # {ru.STATE: None                           , ru.EVENT: 'app_start'      },
    # {ru.STATE: None                           , ru.EVENT: 'app_stop'       },
    # {ru.STATE: None                           , ru.EVENT: 'exec_ok'        },
      {ru.STATE: None                           , ru.EVENT: 'exec_stop'      },
      {ru.STATE: rp.AGENT_STAGING_OUTPUT_PENDING, ru.EVENT: 'state'          },
      {ru.STATE: rp.AGENT_STAGING_OUTPUT        , ru.EVENT: 'state'          },
    # {ru.STATE: rp.TMGR_STAGING_OUTPUT_PENDING , ru.EVENT: 'state'          },
    # {ru.STATE: rp.TMGR_STAGING_OUTPUT         , ru.EVENT: 'state'          },
    # {ru.STATE: rp.DONE                        , ru.EVENT: 'state'          },
]


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    if len(sys.argv) < 2:
        print("\n\tusage: %s <dir|tarball>\n" % sys.argv[0])
        sys.exit(1)

    src     = sys.argv[1]
    stype   = 'radical.pilot'
    session = ra.Session.create(src, stype)

    data = dict()
    pipe = dict()

    for thing in session.get(etype=event_entities):

        tstamps = list()

        for event in event_list:
            times = thing.timestamps(event=event)
            if times: tstamps.append(times[0])
            else    : tstamps.append(None)

        data[thing.uid] = tstamps
        try:
            pipe[thing.uid] = [int(x) for x
                               in thing.description.get('name', '').split()]
        except:
            pipe[thing.uid] = [0]

  # diffs = list()
  # for uid in data:
  #     diffs.append(data[uid][-1] - data[uid][0])
  # print(sorted(diffs))


  # sort x-axis (task IDs) by
  #     'uid'  : task ID
  #     'get'  : time of ingest
  #     'sched': time of scheduling
  #     'pipe' : pipeline ID (RE sessions only)
    order = 'sched'
    index = 4

    if order == 'uid':
        sorted_uids = sorted(pipe.keys())
    else:
        sorted_uids = [x[0] for x in sorted(list(data.items()),
                                            key=lambda v: v[1][index])]
    sorted_data   = list()
    index         = 0
    for uid in sorted_uids:
        sorted_data.append([index] + data[uid])
        index += 1

    np_data = np.array(sorted_data)

    fig, ax = plt.subplots(figsize=ra.get_plotsize(500))

    for e_idx in range(len(event_list)):
        ax.plot(np_data[:,0], np_data[:,(1 + e_idx)],
                 label=to_latex(ru.event_to_label(event_list[e_idx])))

    plt.xlabel(to_latex('task (sorted by %s)' % order))
    plt.ylabel(to_latex('time [sec]'))

    # FIXME: how to do the legend now?  With the large font size, I don't see
    plt.legend()
    plt.savefig('%s_state.png' % session.uid)
  # plt.show()


# ------------------------------------------------------------------------------

